import unittest
import asyncio
from finbourne_sdk_utils.extract.group_holdings import get_holdings_for_group
from pathlib import Path
from finbourne_sdk_utils.cocoon.utilities import create_scope_id
from parameterized import parameterized

from finbourne_sdk_utils import logger
import lusid
import pandas as pd
from finbourne_sdk_utils import cocoon as cocoon
import os
import sys
from lusid.models import CurrencyAndAmount

# Create the Portfolios, Portfolio Groups and Holdings
scope = create_scope_id()


class CocoonTestsExtractGroupHoldings(unittest.IsolatedAsyncioTestCase):
    
 
    @classmethod
    async def setup(cls) -> None:
        
        secrets_file = Path(__file__).parent.parent.parent.joinpath("secrets.json")
        
        
        loop = asyncio.get_event_loop()
                
       # executor = ProcessPoolExecutor(max_workers=worker_count)
        cls.api_factory = lusid.SyncApiClientFactory( )
   
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

        cls.scope = scope

        portfolio_holdings = pd.read_csv(
            Path(__file__).parent.joinpath("./data/holdings-example.csv")
        )
        portfolio_groups = pd.read_csv(
            Path(__file__).parent.joinpath("./data/portfolio-groups.csv")
        )

        # Create portfolios
        response = cocoon.load_from_data_frame(
            api_factory=cls.api_factory,
            scope=cls.scope,
            data_frame=portfolio_holdings,
            mapping_required={
                "code": "FundCode",
                "display_name": "FundCode",
                "created": "$2010-10-09T08:00:00Z",
                "base_currency": "$AUD",
            },
            mapping_optional={},
            file_type="portfolios",
            property_columns=[],
        )
        assert len(response["portfolios"]["errors"]) == 0

        # Add holdings
        response = cocoon.load_from_data_frame(
            api_factory=cls.api_factory,
            scope=cls.scope,
            data_frame=portfolio_holdings,
            mapping_required={
                "code": "FundCode",
                "effective_at": "Effective Date",
                "tax_lots.units": "Quantity",
            },
            mapping_optional={
                "tax_lots.cost.amount": "Local Market Value",
                "tax_lots.cost.currency": "Local Currency Code",
                "tax_lots.portfolio_cost": None,
                "tax_lots.price": None,
                "tax_lots.purchase_date": None,
                "tax_lots.settlement_date": None,
            },
            file_type="holdings",
            identifier_mapping={
                "ClientInternal": "SEDOL Security Identifier",
                "Currency": "is_cash_with_currency",
            },
            property_columns=[],
            holdings_adjustment_only=True,
        )
        assert len(response["holdings"]["errors"]) == 0, len(
            response["holdings"]["errors"]
        )

        # Create groups
        response = cocoon.load_from_data_frame(
            api_factory=cls.api_factory,
            scope=cls.scope,
            data_frame=portfolio_groups,
            mapping_required={
                "code": "PortGroupCode",
                "display_name": "PortGroupDisplayName",
                "created": "$2010-10-09T08:00:00Z",
            },
            mapping_optional={
                "values.scope": f"${cls.scope}",
                "values.code": "FundCode",
            },
            file_type="portfolio_group",
            property_columns=[],
        )
        assert len(response["portfolio_groups"]["errors"]) == 0

        # Create group with sub-groups
        response = cls.api_factory.build(
            lusid.api.PortfolioGroupsApi
        ).create_portfolio_group(
            scope=cls.scope,
            create_portfolio_group_request=lusid.models.CreatePortfolioGroupRequest(
                code="SubGroups",
                display_name="SubGroups",
                created="2010-10-09T08:00:00Z",
                values=[lusid.models.ResourceId(scope=cls.scope, code="Portfolio-Y")],
                sub_groups=[lusid.models.ResourceId(scope=cls.scope, code="ABC12")],
            ),
        )

        assert isinstance(response, lusid.models.PortfolioGroup)


    @classmethod
    def setUpClass(cls) -> None:
        
        
        loop = asyncio.get_event_loop()
        
        
        loop.run_until_complete( cls.setup() )
    

    def check_holdings_correct(
        self, group_holdings, expected_values, lusid_results_keyed
    ):

        # Check that there are the right number of results and they have the correct keys
        self.assertSetEqual(set(group_holdings.keys()), set(expected_values.keys()))

        if lusid_results_keyed is not None:
            self.assertSetEqual(
                set(group_holdings.keys()), set(lusid_results_keyed.keys())
            )

        # Iterate over each result
        for portfolio, holdings in group_holdings.items():

            # Key the result by the instrument name
            holdings_by_instrument_name = {
                holding.properties["Instrument/default/Name"].value.label_value: holding
                for holding in holdings
            }

            # Check that the units and cost are correct according to manual test data
            for instrument_name, holding in holdings_by_instrument_name.items():
                self.assertEqual(
                    float(holding.cost.amount),
                    float(expected_values[portfolio][instrument_name]["cost"]),
                )
                self.assertEqual(
                    float(holding.units),
                    float(expected_values[portfolio][instrument_name]["units"]),
                )
                self.assertEqual(
                    holding.cost.currency,
                    expected_values[portfolio][instrument_name]["cost.currency"],
                )

                # If this is not grouped by Portfolio, skip checking against LUSID as the merging would have to be
                # done above the API anyway
                if lusid_results_keyed is None:
                    return

                # Check that the units and cost are correct according to LUSID results
                self.assertEqual(
                    float(holding.cost.amount),
                    float(lusid_results_keyed[portfolio][instrument_name].cost.amount),
                )
                self.assertEqual(
                    float(holding.units),
                    float(lusid_results_keyed[portfolio][instrument_name].units),
                )
                self.assertEqual(
                    holding.cost.currency,
                    lusid_results_keyed[portfolio][instrument_name].cost.currency,
                )
                self.assertEqual(
                    holding.cost_portfolio_ccy.currency,
                    lusid_results_keyed[portfolio][
                        instrument_name
                    ].cost_portfolio_ccy.currency,
                )
                self.assertEqual(
                    float(holding.cost_portfolio_ccy.amount),
                    float(
                        lusid_results_keyed[portfolio][
                            instrument_name
                        ].cost_portfolio_ccy.amount
                    ),
                )

    @parameterized.expand(
        [
            [
                "Grouped by portfolio, no sub-groups",
                "ABC12",
                True,
                {
                    f"{scope} : Portfolio-X": {
                        "Amazon": {
                            "cost": "52916721",
                            "cost.currency": "USD",
                            "units": "26598",
                        },
                        "Sainsbury": {
                            "cost": "188941977",
                            "cost.currency": "GBP",
                            "units": "942354",
                        },
                        "Tesla Inc": {
                            "cost": "22038434",
                            "cost.currency": "USD",
                            "units": "95421",
                        },
                    },
                    f"{scope} : Portfolio-Z": {
                        "Lloyds Banking Group PLC": {
                            "cost": "141900",
                            "cost.currency": "GBP",
                            "units": "2500",
                        },
                        "Apple Inc": {
                            "cost": "2084000",
                            "cost.currency": "USD",
                            "units": "10000",
                        },
                        "Tesla Inc": {
                            "cost": "22038434",
                            "cost.currency": "USD",
                            "units": "95421",
                        },
                    },
                },
            ],
            [
                "Grouped by portfolio, single sub-group",
                "SubGroups",
                True,
                {
                    f"{scope} : Portfolio-X": {
                        "Amazon": {
                            "cost": "52916721",
                            "cost.currency": "USD",
                            "units": "26598",
                        },
                        "Sainsbury": {
                            "cost": "188941977",
                            "cost.currency": "GBP",
                            "units": "942354",
                        },
                        "Tesla Inc": {
                            "cost": "22038434",
                            "cost.currency": "USD",
                            "units": "95421",
                        },
                    },
                    f"{scope} : Portfolio-Z": {
                        "Lloyds Banking Group PLC": {
                            "cost": "141900",
                            "cost.currency": "GBP",
                            "units": "2500",
                        },
                        "Apple Inc": {
                            "cost": "2084000",
                            "cost.currency": "USD",
                            "units": "10000",
                        },
                        "Tesla Inc": {
                            "cost": "22038434",
                            "cost.currency": "USD",
                            "units": "95421",
                        },
                    },
                    f"{scope} : Portfolio-Y": {
                        "Apple Inc": {
                            "cost": "2084000",
                            "cost.currency": "USD",
                            "units": "10000",
                        },
                        "Amazon": {
                            "cost": "52916721",
                            "cost.currency": "USD",
                            "units": "26598",
                        },
                    },
                },
            ],
            [
                "Merged together, no sub-groups",
                "ABC12",
                False,
                {
                    f"{scope} : ABC12": {
                        "Amazon": {
                            "cost": "52916721",
                            "cost.currency": "USD",
                            "units": "26598",
                        },
                        "Sainsbury": {
                            "cost": "188941977",
                            "cost.currency": "GBP",
                            "units": "942354",
                        },
                        "Tesla Inc": {
                            "cost": "44076868",
                            "cost.currency": "USD",
                            "units": "190842",
                        },
                        "Lloyds Banking Group PLC": {
                            "cost": "141900",
                            "cost.currency": "GBP",
                            "units": "2500",
                        },
                        "Apple Inc": {
                            "cost": "2084000",
                            "cost.currency": "USD",
                            "units": "10000",
                        },
                    }
                },
            ],
            [
                "Merged together, single sub-group",
                "SubGroups",
                False,
                {
                    f"{scope} : SubGroups": {
                        "Amazon": {
                            "cost": "105833442",
                            "cost.currency": "USD",
                            "units": "53196",
                        },
                        "Sainsbury": {
                            "cost": "188941977",
                            "cost.currency": "GBP",
                            "units": "942354",
                        },
                        "Tesla Inc": {
                            "cost": "44076868",
                            "cost.currency": "USD",
                            "units": "190842",
                        },
                        "Lloyds Banking Group PLC": {
                            "cost": "141900",
                            "cost.currency": "GBP",
                            "units": "2500",
                        },
                        "Apple Inc": {
                            "cost": "4168000",
                            "cost.currency": "USD",
                            "units": "20000",
                        },
                    }
                },
            ],
        ]
    )
    def test_get_holdings_for_group_grouped_by_portfolio(
        self, _, group_code, group_by_portfolio, expected_results
    ) -> None:

        # Get the group holdings
        group_holdings = get_holdings_for_group(
            api_factory=self.api_factory,
            group_scope=scope,
            group_code=group_code,
            property_keys=["Instrument/default/Name"],
            group_by_portfolio=group_by_portfolio,
            num_threads=15,
        )

        # If not grouping by Portfolio there is no point checking against LUSID as there is no equivalent
        if not group_by_portfolio:
            # Still check against manually prepared data
            return self.check_holdings_correct(group_holdings, expected_results, None)

        # Otherwise get the result from LUSID
        lusid_results = self.api_factory.build(
            lusid.api.PortfolioGroupsApi
        ).get_holdings_for_portfolio_group(
            scope=self.scope, code=group_code, property_keys=["Instrument/default/Name"]
        )

        # Key the LUSID result against portfolio and then instrument name
        lusid_results_keyed = {}
        for holding in lusid_results.values:
            portfolio_scope = holding.properties[
                "Holding/default/SourcePortfolioScope"
            ].value.label_value
            portfolio_code = holding.properties[
                "Holding/default/SourcePortfolioId"
            ].value.label_value
            portfolio = f"{portfolio_scope} : {portfolio_code}"

            instrument_name = holding.properties[
                "Instrument/default/Name"
            ].value.label_value
            lusid_results_keyed.setdefault(portfolio, {}).setdefault(
                instrument_name, holding
            )

        # Check that the result is as expected
        self.check_holdings_correct(
            group_holdings, expected_results, lusid_results_keyed
        )



import os
import unittest
from pathlib import Path
import pandas as pd
import lusid


from finbourne_sdk_utils import cocoon as cocoon
from finbourne_sdk_utils import logger
from finbourne_sdk_utils.cocoon.utilities import create_scope_id
from parameterized import parameterized


class CocoonTestsReferencePortfolios(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:

        cls.api_factory = lusid.SyncApiClientFactory()
        
        cls.portfolios_api = cls.api_factory.build(lusid.api.PortfoliosApi)
        cls.unique_id = create_scope_id()
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))
        cls.scope = "ModelFundTest"
        cls.file_name = "data/reference-portfolio/reference-test.csv"

    @parameterized.expand(
        [
            [
                "Load a reference portfolio with required attributes only",
                "data/reference-portfolio/reference-test.csv",
                {"code": "FundCode", "display_name": "display_name"},
                {
                    "description": "description",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                [],
            ],
            [
                "Load a reference portfolio with required attributes AND properties",
                "data/reference-portfolio/reference-test.csv",
                {"code": "FundCode", "display_name": "display_name"},
                {
                    "description": "description",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                ["strategy", "custodian"],
            ],
            [
                "Load duplicate reference portfolios",
                "data/reference-portfolio/reference-test-duplicates.csv",
                {"code": "FundCode", "display_name": "display_name"},
                {
                    "description": "description",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                [],
            ],
        ]
    )
    def test_load_from_data_frame_attributes_and_properties_success(
        self, _, file_name, mapping_required, mapping_optional, property_columns,
    ) -> None:
        """
        Test that a reference portfolio can be loaded successfully

        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param list[str] property_columns: The columns to create properties for
        :param any expected_outcome: The expected outcome

        :return: None
        """

        unique_id = create_scope_id()
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))
        data_frame.drop_duplicates(inplace=True)
        data_frame["FundCode"] = data_frame["FundCode"] + "-" + unique_id

        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=self.scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="reference_portfolio",
            identifier_mapping={},
            property_columns=property_columns,
            properties_scope=self.scope,
            sub_holding_keys=[],
        )

        # Check that the count of portfolios uploaded equals count of portfolios in DataFrame

        self.assertEqual(
            first=len(responses["reference_portfolios"]["success"]),
            second=len(data_frame),
        )

        # Assert that by no unmatched_identifiers are returned in the response for reference_portfolios
        self.assertFalse(
            responses["reference_portfolios"].get("unmatched_identifiers", False)
        )

        # Check that the portfolio IDs of porfolios uploaded matches the IDs of portfolios in DataFrame

        response_codes = [
            response.id.code
            if isinstance(response, lusid.models.Portfolio)
            else response.origin_portfolio_id.code
            for response in responses["reference_portfolios"]["success"]
        ]

        self.assertEqual(
            first=response_codes,
            second=list(data_frame[mapping_required["code"]].values),
        )

        # Check that properties get added to portfolio

        for portfolio in responses["reference_portfolios"]["success"]:

            get_portfolio = self.portfolios_api.get_portfolio(
                scope=portfolio.id.scope,
                code=portfolio.id.code,
                property_keys=[
                    f"Portfolio/{self.scope}/strategy",
                    f"Portfolio/{self.scope}/custodian",
                ],
            )

            property_keys_from_params = [
                f"Portfolio/{self.scope}/{code}"
                for code in property_columns
                if len(property_columns) > 0
            ]

            self.assertCountEqual(
                [prop for prop in get_portfolio.properties], property_keys_from_params,
            )

    def test_portfolio_missing_attribute(self):

        unique_id = create_scope_id()
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(self.file_name))
        data_frame["FundCode"] = data_frame["FundCode"] + "-" + unique_id

        mapping_required = {"code": "FundCode"}
        mapping_optional = {
            "description": "description",
            "created": "created",
            "base_currency": "base_currency",
        }

        with self.assertRaises(ValueError) as error:

            cocoon.cocoon.load_from_data_frame(
                api_factory=self.api_factory,
                scope=self.scope,
                data_frame=data_frame,
                mapping_required=mapping_required,
                mapping_optional=mapping_optional,
                file_type="reference_portfolio",
                identifier_mapping={},
                property_columns=[],
                properties_scope=self.scope,
                sub_holding_keys=[],
            )

        self.assertEqual(
            error.exception.args[0],
            """The required attributes {'display_name'} are missing from the mapping. Please
                             add them.""",
        )



from typing import List

from finbourne_sdk_utils.cocoon.utilities import checkargs
from finbourne_sdk_utils import cocoon
import lusid
import pandas as pd
import logging
from http import HTTPStatus
import numpy as np

# Map Numpy data types to LUSID data types
global_constants = {
    "data_type_mapping": {
        "object": "string",
        "float64": "number",
        "int64": "number",
        "bool": "string",
        "datetime64[ns, UTC]": "string",
    }
}


@checkargs
def check_property_definitions_exist_in_scope_single(
    api_factory: lusid.SyncApiClientFactory, property_key: str
) -> tuple[bool, str]:
    """
    This function takes a list of property keys and looks to see which property definitions already exist inside LUSID

    Parameters
    ----------
    api_factory : lusid.SyncApiClientFactory
        The ApiFactory to use
    property_key: str
        The property key to get from LUSID

    Returns
    -------
    exists : bool
        Whether or not the property definition exists
    data_type : str
        property definition's data type
    """

    data_type = None

    try:
        response = api_factory.build(
            lusid.PropertyDefinitionsApi
        ).get_property_definition(
            domain=property_key.split("/")[0],
            scope=property_key.split("/")[1],
            code=property_key.split("/")[2],
        )

        exists = True
        data_type = response.data_type_id.code

    except lusid.exceptions.ApiException as ex:
        if ex.status == HTTPStatus.NOT_FOUND:
            exists = False
        else:
            raise

    return exists, data_type


@checkargs
def check_property_definitions_exist_in_scope(
    api_factory: lusid.SyncApiClientFactory,
    domain: str,
    data_frame: pd.DataFrame,
    target_columns: list,
    column_to_scope: dict,
):
    """
    This function identifiers which property definitions are missing from LUSID

    Parameters
    ----------
    api_factory :   lusid.SyncApiClientFactory
        The Api Factory to use
    domain : str
        The domain to check for property definitions in
    data_frame : pd.DataFrame
        The dataframe to check properties for
    target_columns : list[str]
        The columns to add properties for
    column_to_scope : dict[str:str]
        Column name to scope
    Returns
    -------
    missing_property_columns :  list[str]
        The columns missing properties in LUSID
    data_frame : pd.DataFrame
        The input DataFrame with types updated
    """

    data_type_update_map = {"number": "float64", "string": "object"}

    # Initialise a set to hold the missing properties
    missing_keys = set([])

    # Iterate over the column names
    column_property_mapping = {}

    for column_name, data_type in data_frame.loc[:, target_columns].dtypes.items():

        # Create the property key
        property_key = f"{domain}/{column_to_scope[column_name]}/{cocoon.utilities.make_code_lusid_friendly(column_name)}"

        column_property_mapping[property_key] = column_name

        # Get a tuple with the first value being True/False key is missing, second is data type of the key
        exists, data_type_lusid = check_property_definitions_exist_in_scope_single(
            api_factory=api_factory, property_key=property_key
        )

        # If the key is missing add it to the set
        if not exists:
            missing_keys.add(property_key)

        # If it is not missing check that the data type of the property matches the dataframe
        else:
            # If the data type does not match
            if data_type_lusid != global_constants["data_type_mapping"][str(data_type)]:
                logging.warning(
                    f"Data types don't match for column {column_name} it is {data_type_lusid} in LUSID and {data_type} in file"
                )
                try:
                    updated_data_type = data_type_update_map[data_type_lusid]
                except KeyError:
                    updated_data_type = "object"

                # Update the data type in the dataframe if possible
                data_frame[column_name] = data_frame[column_name].astype(
                    updated_data_type, copy=False
                )

                logging.info(f"Updated {column_name} to {updated_data_type}")

    missing_property_columns = [
        column_property_mapping[property_key] for property_key in missing_keys
    ]

    return missing_property_columns, data_frame


@checkargs
def create_property_definitions_from_file(
    api_factory: lusid.SyncApiClientFactory,
    domain: str,
    data_frame: pd.DataFrame,
    missing_property_columns: list,
    column_to_scope: dict,
):
    """
    Creates the property definitions for all the columns in a file

    Parameters
    ----------
    api_factory : lusid.SyncApiClientFactory
        The ApiFactory to use
    domain : domain
        The domain to create the property definitions in
    data_frame : pd.Series
        The dataframe dtypes to add definitions for
    missing_property_columns : list[str]
        The columns that property defintions are missing for
    column_to_scope : dict[str:str]
        Column name to scope

    Returns
    -------
    property_key_mapping : dict
        A mapping of data_frame columns to property keys
    """

    missing_property_data_frame = data_frame.loc[:, missing_property_columns]

    # Ensure that all data types in the file have been mapped
    actual_data_types = set(
        [str(data_type) for data_type in missing_property_data_frame.dtypes]
    )
    allowed_data_types = set(global_constants["data_type_mapping"])
    if not (actual_data_types <= allowed_data_types):
        unmapped_data_types = [
            np.dtype(value) for value in actual_data_types - allowed_data_types
        ]
        unmapped_columns = missing_property_data_frame.dtypes[
            missing_property_data_frame.dtypes.isin(unmapped_data_types)
        ]
        raise TypeError(
            invalid_columns_error_message(unmapped_columns, allowed_data_types)
        )

    # Initialise a dictionary to hold the keys
    property_key_mapping = {}

    # Iterate over the each column and its data type
    for column_name, data_type in missing_property_data_frame.dtypes.items():

        # Make the column name LUSID friendly
        lusid_friendly_code = cocoon.utilities.make_code_lusid_friendly(column_name)

        # Check if column is all nulls
        if missing_property_data_frame[column_name].isnull().all():
            logging.warning(
                f"{column_name} is entirely null, no type can be inferred - will be treated as a string"
            )
            data_type = "object"
            data_frame[column_name] = data_frame[column_name].astype(
                "object", copy=False
            )
        else:
            # Find the first non-null value to infer the proper type
            first_valid_value = missing_property_data_frame[column_name].dropna().iloc[0]
            
            # Infer the type from the first non-null value
            inferred_type = type(first_valid_value).__name__
            
            # Map Python types to pandas dtypes
            type_mapping = {
                'str': 'object',
                'int': 'int64',
                'float': 'float64',
                'bool': 'bool',
                'Timestamp': 'datetime64[ns]',
            }
            
            # Get the appropriate pandas dtype
            inferred_dtype = type_mapping.get(inferred_type, 'object')
            
            # Only override if the current dtype is object and we found something more specific
            if str(data_type) == 'object' and inferred_dtype != 'object':
                logging.info(
                    f"{column_name} has leading nulls - inferring type '{inferred_dtype}' from first non-null value"
                )
                data_type = inferred_dtype
                # Optionally convert the column to the inferred type
                try:
                    data_frame[column_name] = data_frame[column_name].astype(
                        inferred_dtype, copy=False
                    )
                except (ValueError, TypeError) as e:
                    logging.warning(
                        f"Could not convert {column_name} to {inferred_dtype}, keeping as object. Error: {e}"
                    )
                    data_type = 'object'

        # Create a request to define the property, assumes value_required is false for all
        property_request = lusid.models.CreatePropertyDefinitionRequest(
            domain=domain,
            scope=column_to_scope[column_name],
            code=lusid_friendly_code,
            value_required=False,
            display_name=column_name,
            data_type_id=lusid.models.ResourceId(
                scope="system",
                code=global_constants["data_type_mapping"][str(data_type)],
            ),
        )

        # Call LUSID to create the new property
        property_response = api_factory.build(
            lusid.PropertyDefinitionsApi
        ).create_property_definition(
            create_property_definition_request=property_request
        )

        logging.info(
            f"Created - {property_response.key} - with datatype {property_response.data_type_id.code}"
        )

        # Grab the key off the response to use when referencing this property in other LUSID calls
        property_key_mapping[column_name] = property_response.key

    return property_key_mapping, data_frame


@checkargs
def create_missing_property_definitions_from_file(
    api_factory: lusid.SyncApiClientFactory,
    properties_scope: str,
    data_frame: pd.DataFrame,
    property_columns: list,
    domain: str,
):
    # If there are property columns
    if len(property_columns) > 0 and domain is not None:

        source_columns = [column["source"] for column in property_columns]
        source_to_target = {
            column1["source"]: column1.get("target", column1.get("source"))
            for column1 in property_columns
        }

        for column in source_columns:
            data_frame.loc[:, source_to_target[column]] = data_frame[column]

        target_columns = [
            column.get("target", column.get("source")) for column in property_columns
        ]
        column_to_scope = {
            column.get("target", column.get("source")): column.get(
                "scope", properties_scope
            )
            for column in property_columns
        }

        # Identify which property definitions are missing
        (
            missing_property_columns,
            data_frame,
        ) = cocoon.properties.check_property_definitions_exist_in_scope(
            api_factory=api_factory,
            domain=domain,
            data_frame=data_frame,
            target_columns=target_columns,
            column_to_scope=column_to_scope,
        )

        logging.info(
            f"Check for missing {domain} properties complete. {len(missing_property_columns)} missing properties found"
        )

        # If there are missing property definitions
        if len(missing_property_columns) > 0:
            logging.info(
                f"The {domain} properties {str(missing_property_columns)} will be added in the scope {properties_scope}"
            )

            # Create property definitions for all of the columns in the file that have missing definitions
            (
                property_key_mapping,
                data_frame,
            ) = cocoon.properties.create_property_definitions_from_file(
                api_factory=api_factory,
                domain=domain,
                data_frame=data_frame,
                missing_property_columns=missing_property_columns,
                column_to_scope=column_to_scope,
            )

    return data_frame


def invalid_columns_error_message(unmapped_columns, allowed_data_types):
    formatted_unmapped_columns = {
        k: str(v) for k, v in unmapped_columns.to_dict().items()
    }
    return f"""The following columns in the data_frame have not been mapped to LUSID data types: {formatted_unmapped_columns}.
                           LUSID supports the following data types: {allowed_data_types}. 
                           Please ensure that all data types have been mapped before retrying."""


@checkargs
def create_property_values(
    row: pd.Series, column_to_scope: dict, scope: str, domain: str, dtypes: pd.Series
) -> dict:
    """
    This function generates the property values for a row in a file

    Parameters
    ----------
    row : pd.Series
        The current row of the data frame to create property values for
    column_to_scope : dict {str, str}
        The scope for a column name
    scope : str
        The domain to create the property values in
    domain : str
        The domain to create the property values in
    dtypes : pd.Series
        The data types of each column to create property values for

    Returns
    -------
    properties : dict {str, models.PerpetualProperty}
    """

    actual_data_types = set([str(data_type) for data_type in dtypes])
    allowed_data_types = set(global_constants["data_type_mapping"])

    # Ensure that all data types in the file have been mapped
    if not (actual_data_types <= allowed_data_types):
        unmapped_data_types = actual_data_types - allowed_data_types
        unmapped_columns = dtypes[dtypes.isin(unmapped_data_types)]
        raise TypeError(
            invalid_columns_error_message(unmapped_columns, allowed_data_types)
        )

    # Initialise the empty properties dictionary
    properties = {}

    # Iterate over each column name and data type
    for column_name, data_type in dtypes.items():

        # Set the data type to be a string so that it is easier to work with
        string_data_type = str(data_type)
        # Convert the numpy data type to a LUSID data type using the global mapping
        lusid_data_type = global_constants["data_type_mapping"][string_data_type]
        # Get the value of the column from the row
        row_value = row[column_name]

        # Use the correct LUSID property value based on the data type
        if lusid_data_type == "string":
            if pd.isna(row_value):
                continue
            property_value = lusid.models.PropertyValue(label_value=str(row_value))  

        if lusid_data_type == "number":
            # Handle null values given the input null value override
            if pd.isnull(row_value):
                continue
            property_value = lusid.models.PropertyValue(
                metric_value=lusid.models.MetricValue(value=row_value)
            )

        # Set the property
        property_key = f"{domain}/{column_to_scope.get(column_name, scope)}/{cocoon.utilities.make_code_lusid_friendly(column_name)}"
        properties[property_key] = lusid.models.PerpetualProperty(
            key=property_key, value=property_value
        )

    if domain.lower() == "instrument":
        properties = list(properties.values())

    return properties


def _infer_full_property_keys(
    partial_keys: list, properties_scope: str, domain: str
) -> list:
    """
    Infers from a list of partially completed property keys the entire property key

    Parameters
    ----------
    partial_keys : list[str]
        The partial keys
    properties_scope : str
        The scope of the properties
    domain : str
        The domain of the properties
    Returns
    -------
    list[str]
        A list of full property keys
    """

    full_keys = []

    for key in partial_keys:

        split_key = key.split("/")
        number_components = len(split_key)
        # The entire key is already specified
        if number_components == 3:
            full_keys.append(key)
        # The domain is missing with the scope and code specified
        elif number_components == 2:
            full_keys.append(f"{domain}/{split_key[0]}/{split_key[1]}")
        # The domain and scope are missing with only the code specified
        elif number_components == 1:
            full_keys.append(f"{domain}/{properties_scope}/{key}")

    # Ensure that the returned keys are LUSID friendly
    return [
        "/".join(
            [
                cocoon.utilities.make_code_lusid_friendly(partial)
                for partial in key.split("/")
            ]
        )
        for key in full_keys
    ]



import argparse

# Ensure standardisation of commonly used arguments


class Parser:

    # Create a parser and add in the standard arguments
    def __init__(self, description, sections=[]):
        self.parser = argparse.ArgumentParser(
            description=description, fromfile_prefix_chars="@"
        )
        self.post_processors = []
        self.arguments = []

        if "scope" in sections:
            self.add("scope", help="Scope")

        if "portfolio" in sections:
            self.add("portfolio", help="Portfolio id")

        if "date" in sections:
            self.add("date", help="date YYYY-MM-DD")

        if "input" in sections:
            self.add("input", help="input filename")

        if "properties" in sections:
            self.add("--properties", nargs="+", help="List of propertykeys", default=[])

        if "filename" in sections:
            self.add(
                "-f", "--filename", metavar="filename.csv", help="write to this file"
            )

        if "limit" in sections:
            self.add(
                "-l",
                "--limit",
                type=int,
                default=0,
                metavar="n",
                help="limit the number of results",
            )

        if "date_range" in sections:
            self.add("-s", "--start_date", dest="start_date", metavar="YYYY-MM-DD")
            self.add("-e", "--end_date", dest="end_date", metavar="YYYY-MM-DD")

        if "quiet" in sections:
            self.add(
                "-q",
                "--quiet",
                action="store_true",
                help="Quiet mode. Doesn't show the progress bar",
            )

        if "group" in sections:
            self.add(
                "-g",
                "--group",
                action="store_true",
                help="Indicates use of Portfolio Groups",
            )

        if "optional_portfolio" in sections:
            self.add(
                "-p",
                "--portfolio",
                dest="portfolio",
                help="Optional Portfolio id"
            )

        if "flush_scope" in sections:
            self.add(
                "--flush_scope",
                dest="flush_scope",
                action="store_true",
                help="Flush all transactions in scope"
            )

        self.add(
            "--secrets-file",
            dest="secrets",
            default="secrets.json",
            help="path to secrets file",
        )

        self.add(
            "--environment",
            dest="env",
            default=["lusid"],
            nargs="+",
            help="choose a special LUSID environment. E.g. 'ipsum_lorem'",
        )

        self.add(
            "--stats",
            dest="stats",
            metavar="stats-file.csv",
            const="-",
            nargs="?",
            help="Write statistics to file. [Leave blank for stdout]",
        )

        self.add("-d", "--debug", help=r"print debug messages, expected input: 'debug'")

        if "asat" in sections:
            self.add(
                "--asat",
                dest="asat",
                metavar="YYYY-MM-DDTHH:MM:SS.000",
                help="as-at time",
            )

        if "test" in sections:
            self.add("--test", action="store_true", help="Run in test mode")

        # If a filename parameter is given then pass output via dfq
        # Can be supressed by the NODFQ option
        if "filename" in sections and "NODFQ" not in sections:
            self.add(
                "--dfq",
                nargs=argparse.REMAINDER,
                help="pass the output through to 'dfq' - see `finbourne_sdk_utils dfq --help` for options",
            )

    def add(self, *args, **kwargs):
        # Add arguments to a list, so they can be removed
        self.arguments.append((args[0], args, kwargs))
        # self.parser.add_argument(*args,**kwargs)
        return self

    def remove(self, key):
        self.arguments = [tpl for tpl in self.arguments if tpl[0] != key]

    def post_process(self, fn):
        self.post_processors.append(fn)
        return self

    def parse(self, args=None):
        for arg in self.arguments:
            self.parser.add_argument(*arg[1], **arg[2])

        args = self.parser.parse_args(args)
        for fn in self.post_processors:
            fn(args)

        return args

    def extend(self, fn):
        if fn:
            fn(self)
        return self



class Either:
    def __init__(self, *args, **kwargs):

        # at least 2 arguments. Assume first two are Left and Right
        if len(args) >= 2:
            self.left = args[0]
            self.right = args[1]
            # 1 argument. May be 'dict like'
        elif len(args) == 1:
            if isinstance(args[0], Either):
                self.left = args[0].left
                self.right = args[0].right
            else:
                try:
                    self.right = args[0].get("right")
                    self.left = args[0].get("left")
                except:
                    try:
                        self.right = args[0].right
                        self.left = args[0].left
                    except:
                        pass

                if self.left is None and self.right is None:
                    # only 1 value, Just assume it's a 'right'
                    self.right = args[0]

        # No positional arguments - could be provided via kwargs
        else:
            self.left = kwargs.get("left", None)
            self.right = kwargs.get("right", None)

    # Call the left function if left is available
    # Otherwise call the right function
    def match(self, left, right):
        if self.left is not None:
            return left(self.left)

        return right(self.right)

    def is_left(self):
        return self.left is not None

    def if_left(self, left):
        if self.left is not None:
            return left(self.left)

    def is_right(self):
        return self.right is not None

    def if_right(self, right):
        if self.right is not None:
            return right(self.right)

    def __getattr__(self, name):
        return self.__dict__.get(name)

    def bind(self, fn):
        return Either(fn(self.right)) if self.left is None else self

    def Left(v):
        return Either(left=v)

    def Right(v):
        return Either(right=v)



import os
import unittest
from pathlib import Path
import pandas as pd

from finbourne_sdk_utils import cocoon as cocoon
from parameterized import parameterized
import lusid
from finbourne_sdk_utils import logger
from datetime import datetime
import pytz
from finbourne_sdk_utils.cocoon.utilities import create_scope_id

unique_properties_scope = create_scope_id()


def expected_response(property_scope="TestPropertiesScope1"):
    return {
        "instruments": lusid.models.UpsertInstrumentsResponse(
            values={
                "ClientInternal: imd_34534539": lusid.models.Instrument(
                    lusid_instrument_id="LUIDUnknown",
                    version=lusid.models.Version(
                        as_at_date=datetime.now(pytz.UTC),
                        effective_from=datetime.now(pytz.UTC),
                    ),
                    name="USTreasury_6.875_2025",
                    identifiers={
                        "ClientInternal": "imd_34534539",
                        "Figi": "BBG000DQQNJ8",
                        "Isin": "US912810EV62",
                    },
                    state="Active",
                    properties=[
                        lusid.models.ModelProperty(
                            effective_from=datetime.min.replace(tzinfo=pytz.UTC),
                            effective_until=datetime.max.replace(tzinfo=pytz.UTC),
                            key=f"Instrument/{property_scope}/currency",
                            value=lusid.models.PropertyValue(label_value="USD"),
                        ),
                        lusid.models.ModelProperty(
                            effective_from=datetime.min.replace(tzinfo=pytz.UTC),
                            effective_until=datetime.max.replace(tzinfo=pytz.UTC),
                            key=f"Instrument/{property_scope}/moodys_rating",
                            value=lusid.models.PropertyValue(label_value="Aa2"),
                        ),
                    ],
                )
            },
            failed={},
        )
    }


class CocoonTestsInstruments(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:

        cls.api_factory = lusid.SyncApiClientFactory()
        
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    @parameterized.expand(
        [
            [
                "A standard successful load of scoped instruments",
                "TestScope1",
                "data/global-fund-combined-instrument-master.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "TestScope1",
                expected_response(),
            ],
            [
                "A successful load of instruments to default scope",
                "TestScope1",
                "data/global-fund-combined-instrument-master.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                None,
                expected_response(),
            ],
            [
                "A standard successful load of instruments with string index",
                "TestScope1",
                "data/global-fund-combined-instrument-master-string-index.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "TestScope1",
                expected_response(),
            ],
            [
                "A standard successful load of instruments with duplicates in index",
                "TestScope1",
                "data/global-fund-combined-instrument-master-duplicate-index.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "TestScope1",
                expected_response(),
            ],
            [
                "A standard successful load of instruments with unique properties scope",
                "TestScope1",
                "data/global-fund-combined-instrument-master.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                f"TestPropertiesScope1_{unique_properties_scope}",
                "TestScope1",
                expected_response(f"TestPropertiesScope1_{unique_properties_scope}"),
            ],
            [
                "Additional attributes in the both required and optional mapping that don't exist in LUSID but are in the dataframe",
                "TestScope1",
                "data/global-fund-combined-instrument-master.csv",
                {"name": "instrument_name", "position": "instrument_name"},
                {"transaction.cost": "currency"},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "TestScope1",
                expected_response(),
            ],
            [
                "A different way of specifying the identifiers",
                "TestScope1",
                "data/global-fund-combined-instrument-master.csv",
                {"name": "instrument_name"},
                {},
                {
                    "Instrument/default/Figi": "figi",
                    "Isin": "isin",
                    "Instrument/default/ClientInternal": "client_internal",
                },
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "TestScope1",
                expected_response(),
            ],
            [
                "Providing a default for a required field which has some missing values",
                "TestScope1",
                "data/global-fund-combined-instrument-master-missing-names.csv",
                {
                    "name": {
                        "column": "instrument_name",
                        "default": "<NameCurrentlyUnknown>",
                    }
                },
                {},
                {
                    "Instrument/default/Figi": "figi",
                    "Isin": "isin",
                    "Instrument/default/ClientInternal": "client_internal",
                },
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "TestScope1",
                expected_response(),
            ],
            [
                "Providing a default for a required field which has some missing values",
                "TestScope1",
                "data/global-fund-combined-instrument-master-missing-names.csv",
                {
                    "name": {
                        "column": "instrument_name",
                        "default": "<NameCurrentlyUnknown>",
                    }
                },
                {},
                {
                    "Instrument/default/Figi": "figi",
                    "Isin": "isin",
                    "Instrument/default/ClientInternal": "client_internal",
                },
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "TestScope1",
                expected_response(),
            ],
            [
                "Loading instruments with duplicate rows for the same instrument",
                "TestScope1",
                "data/global-fund-combined-instrument-master-duplicates.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "TestScope1",
                expected_response(),
            ],
        ]
    )
    def test_load_from_data_frame_instruments_success(
        self,
        _,
        scope,
        file_name,
        mapping_required,
        mapping_optional,
        identifier_mapping,
        property_columns,
        properties_scope,
        instrument_scope,
        expected_outcome,
    ) -> None:
        """
        Test that instruments can be loaded successfully

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list[str] property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to
        :param str | None instrument_scope: The scope to add the instruments to
        :param any expected_outcome: The expected outcome

        :return: None
        """

        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="instruments",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            instrument_scope=instrument_scope,
        )

        self.assertEqual(
            0,
            len(responses["instruments"]["errors"]),
            responses["instruments"]["errors"],
        )

        failed = [
            f
            for response in responses["instruments"]["success"]
            for f in response.failed.values()
        ]

        self.assertEqual(0, len(failed), failed)

        self.assertEqual(
            first=sum(
                [
                    len(response.values)
                    for response in responses["instruments"]["success"]
                ]
            ),
            second=len(data_frame["client_internal"].unique()),
        )

        # Assert that by no unmatched_identifiers are returned in the response for instruments
        self.assertFalse(responses["instruments"].get("unmatched_identifiers", False))

        combined_successes = {
            correlation_id: instrument
            for response in responses["instruments"]["success"]
            for correlation_id, instrument in response.values.items()
        }

        self.assertTrue(
            expr=all(
                sorted(combined_successes[key].properties, key=lambda p: p.key)
                == sorted(value.properties, key=lambda p: p.key)
                for key, value in expected_outcome["instruments"].values.items()
            )
        )

        # Assert that instruments has been assigned a scope value
        self.assertTrue(
            expr=all(
                instrument.scope == instrument_scope if instrument_scope is not None else "default"
                for response in responses["instruments"]["success"]
                for instrument in response.values.values()
            )
        )
    @parameterized.expand(
        [
            [
                "Using instrument name enrichment to populate missing names",
                "TestScope1",
                "data/global-fund-combined-instrument-master-missing-names.csv",
                {
                    "name": {
                        "column": "instrument_name",
                        "default": "<NameCurrentlyUnknown>",
                    }
                },
                {},
                {
                    "Instrument/default/Figi": "figi",
                    "Isin": "isin",
                    "Instrument/default/ClientInternal": "client_internal",
                },
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                expected_response(),
            ]
        ]
    )
    @unittest.skip("Skipping test as OpenFigi is broken currently")
    def test_load_from_data_frame_instruments_enrichment_success(
        self,
        _,
        scope,
        file_name,
        mapping_required,
        mapping_optional,
        identifier_mapping,
        property_columns,
        properties_scope,
        expected_outcome,
    ) -> None:
        """
        Test that instruments can be loaded successfully

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list[str] property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to
        :param any expected_outcome: The expected outcome

        :return: None
        """

        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="instruments",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            instrument_name_enrichment=True,
        )

        self.assertEqual(
            first=sum(
                [
                    len(response.values)
                    for response in responses["instruments"]["success"]
                ]
            ),
            second=len(data_frame),
        )

        self.assertEqual(
            first=responses["instruments"]["success"][0]
            .values["ClientInternal: imd_43535553"]
            .name,
            second="BP PLC",
        )

        self.assertEqual(
            first=sum(
                [
                    len(response.failed)
                    for response in responses["instruments"]["success"]
                ]
            ),
            second=0,
        )

        combined_successes = {
            correlation_id: instrument
            for response in responses["instruments"]["success"]
            for correlation_id, instrument in response.values.items()
        }

        self.assertTrue(
            expr=all(
                sorted(combined_successes[key].properties, key=lambda p: p.key)
                == sorted(value.properties, key=lambda p: p.key)
                for key, value in expected_outcome["instruments"].values.items()
            )
        )


    @parameterized.expand(
        [
            [
                "A standard successful load of instruments with whitespace",
                "TestScope2",
                "data/global-fund-combined-instrument-master-with-whitespace.csv",
                "data/global-fund-combined-instrument-master.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                expected_response(),
            ]
        ]
    )
    def test_load_from_data_frame_instruments_with_strip(
        self,
        _,
        scope,
        file_name_with_whitespace,
        file_name_clean,
        mapping_required,
        mapping_optional,
        identifier_mapping,
        property_columns,
        properties_scope,
        expected_outcome,
    ) -> None:
        """
        Test that instruments can be loaded successfully with whitespace, and validates that the

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name_with_whitespace: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list[str] property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to
        :param any expected_outcome: The expected outcome

        :return: None
        """

        data_frame = pd.read_csv(
            Path(__file__).parent.joinpath(file_name_with_whitespace)
        )
        data_frame_true = pd.read_csv(Path(__file__).parent.joinpath(file_name_clean))
        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="instruments",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            remove_white_space=True,
        )

        self.assertEqual(
            len(data_frame_true["client_internal"].unique()),
            sum(
                [
                    len(response.values)
                    for response in responses["instruments"]["success"]
                ]
            ),
        )

        self.assertEqual(
            0,
            sum(
                [
                    len(response.failed)
                    for response in responses["instruments"]["success"]
                ]
            ),
        )

        combined_successes = {
            correlation_id: instrument
            for response in responses["instruments"]["success"]
            for correlation_id, instrument in response.values.items()
        }

        self.assertTrue(
            expr=all(
                sorted(combined_successes[key].properties, key=lambda p: p.key)
                == sorted(value.properties, key=lambda p: p.key)
                for key, value in expected_outcome["instruments"].values.items()
            )
        )

    def test_load_instrument_properties(self,):
        data_frame = pd.DataFrame(
            {
                "instrument_name": [
                    "BP_LondonStockEx_BP",
                    "Glencore_LondonStockEx_GLEN",
                    "TESCO_LondonStockEx_TSCO",
                ],
                "figi": ["BBG000C05BD1", "BBG001MM1KV4", "BBG000BF46Y8"],
                "isin": ["BBG000C05BD1", "JE00B4T3BW64", "GB0008847096"],
                "client_internal": ["imd_43535553", "imd_34534555", "imd_34634673"],
            }
        )

        scope = "TestScope1"
        mapping_required = {"name": "instrument_name"}
        identifier_mapping = {
            "Figi": "figi",
            "Isin": "isin",
            "ClientInternal": "client_internal",
        }

        # set up initial instruments
        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional={},
            file_type="instruments",
            identifier_mapping=identifier_mapping,
        )

        self.assertGreater(len(responses["instruments"]["success"]), 0)

        properties_df = pd.DataFrame(
            {
                "isin": ["BBG000C05BD1", "GG00B4L84979", "GB0031509804"],
                "category": ["Oil & Gas", "Mining", "Retail"],
            }
        )

        properties_required_mapping = {"identifier": "isin", "identifier_type": "$Isin"}

        result = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=properties_df,
            mapping_required=properties_required_mapping,
            mapping_optional={},
            file_type="instrument_property",
            properties_scope=scope,
            property_columns=["category"],
            batch_size=1,
        )

        errors = [
            error
            for batch in result["instrument_propertys"]["errors"]
            for error in batch
        ]
        successes = [
            success
            for batch in result["instrument_propertys"]["success"]
            for success in batch
        ]

        self.assertEqual(len(errors), 0)
        self.assertGreater(len(successes), 0)

        # Assert that by no unmatched_identifiers are returned in the response for instrument_propertys
        self.assertFalse(
            result["instrument_propertys"].get("unmatched_identifiers", False)
        )

    def test_load_instrument_properties_with_missing_instruments(self):
        properties_df = pd.DataFrame({"isin": ["blah"], "category": ["Oil & Gas"]})

        properties_required_mapping = {"identifier": "isin", "identifier_type": "$Isin"}

        scope = "TestScope1"

        result = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=properties_df,
            mapping_required=properties_required_mapping,
            mapping_optional={},
            file_type="instrument_property",
            properties_scope=scope,
            property_columns=["category"],
        )

        errors = [
            error
            for batch in result["instrument_propertys"]["errors"]
            for error in batch
        ]
        successes = [
            success
            for batch in result["instrument_propertys"]["success"]
            for success in batch
        ]

        self.assertEqual(len(errors), 0)
        self.assertEqual(len(successes), 0)

    @parameterized.expand(
        [
            [
                "load only lookthrough instruments",
                Path(__file__).parent.joinpath(
                    "data/lookthrough_instr_tests/load_lookthrough_instrument.csv"
                ),
                {
                    "identifier_mapping": {"ClientInternal": "client_internal",},
                    "required": {"name": "instrument_name"},
                    "optional": {
                        "look_through_portfolio_id.scope": "lookthrough_scope",
                        "look_through_portfolio_id.code": "lookthrough_code",
                    },
                },
            ],
            [
                "load mixed lookthrough instruments",
                Path(__file__).parent.joinpath(
                    "data/lookthrough_instr_tests/mixed_lookthrough_instruments.csv"
                ),
                {
                    "identifier_mapping": {"ClientInternal": "client_internal",},
                    "required": {"name": "instrument_name"},
                    "optional": {
                        "look_through_portfolio_id.scope": "lookthrough_scope",
                        "look_through_portfolio_id.code": "lookthrough_code",
                    },
                },
            ],
            [
                "load mixed lookthrough instruments with default scope",
                Path(__file__).parent.joinpath(
                    "data/lookthrough_instr_tests/mixed_instruments_default_scope.csv"
                ),
                {
                    "identifier_mapping": {"ClientInternal": "client_internal",},
                    "required": {"name": "instrument_name"},
                    "optional": {
                        "look_through_portfolio_id.scope": "$test-lookthrough-loading-finbourne_sdk_utils",
                        "look_through_portfolio_id.code": "lookthrough_code",
                    },
                },
            ],
            [
                "load mixed lookthrough instruments with default scope with multiple portfolios",
                Path(__file__).parent.joinpath(
                    "data/lookthrough_instr_tests/mixed_lookthrough_instruments_multiple portfolios.csv"
                ),
                {
                    "identifier_mapping": {"ClientInternal": "client_internal",},
                    "required": {"name": "instrument_name"},
                    "optional": {
                        "look_through_portfolio_id.scope": "$test-lookthrough-loading-finbourne_sdk_utils",
                        "look_through_portfolio_id.code": "lookthrough_code",
                    },
                },
            ],
            [
                "multiple_instruments_with_same_portfolio",
                Path(__file__).parent.joinpath(
                    "data/lookthrough_instr_tests/multiple_instruments_with_same_portfolio.csv"
                ),
                {
                    "identifier_mapping": {"ClientInternal": "client_internal",},
                    "required": {"name": "instrument_name"},
                    "optional": {
                        "look_through_portfolio_id.scope": "$test-lookthrough-loading-finbourne_sdk_utils",
                        "look_through_portfolio_id.code": "lookthrough_code",
                    },
                },
            ],
        ]
    )
    def test_load_instrument_lookthrough(self, _, df, mapping):

        if "default scope" in _:
            self.skipTest("Default parameter using '$' is not supported")

        scope = "test-lookthrough-loading-finbourne_sdk_utils"
        df = pd.read_csv(df)

        # replace lookthrough scope
        df = df.replace({"replace_scope": scope})

        # populate portfolio ids with random codes
        codes = {
            row["client_internal"]: create_scope_id(use_uuid=True)
            if "Portfolio" in row["instrument_name"]
            else row["client_internal"]
            for index, row in df.iterrows()
        }
        df = df.replace(codes)

        # create portfolios
        [
            self.api_factory.build(lusid.api.TransactionPortfoliosApi).create_portfolio(
                scope=scope,
                create_transaction_portfolio_request=lusid.models.CreateTransactionPortfolioRequest(
                    display_name=row["client_internal"],
                    description=row["client_internal"],
                    code=row["client_internal"],
                    base_currency="USD",
                ),
            )
            if "Portfolio" in row["instrument_name"]
            else None
            for index, row in df.drop_duplicates("client_internal").iterrows()
        ]

        # Upsert lookthrough instrument of portfolio
        instr_response = cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=df,
            mapping_required=mapping["required"],
            mapping_optional=mapping["optional"],
            file_type="instruments",
            identifier_mapping=mapping["identifier_mapping"],
            property_columns=[],
        )

        # check successes, errors and instrument lookthrough codes
        self.assertEqual(
            len(instr_response["instruments"]["success"][0].values.values()), len(df)
        )
        self.assertEqual(len(instr_response["instruments"]["errors"]), 0)

        # check lookthrough code on response
        [
            self.assertEqual(
                instr_response["instruments"]["success"][0]
                .values[f"ClientInternal: {row['client_internal']}"]
                .lookthrough_portfolio.code,
                row["lookthrough_code"],
            )
            if "id" not in row["client_internal"]
            else None
            for index, row in df.iterrows()
        ]

        # tear down this test
        [
            self.api_factory.build(lusid.api.PortfoliosApi).delete_portfolio(
                scope=scope, code=code
            )
            if "id" not in code
            else None
            for code in list(codes.values())
        ]
        [
            self.api_factory.build(lusid.api.InstrumentsApi).delete_instrument(
                "ClientInternal", CI,
            )
            for CI in list(df["client_internal"])
        ]


# Import setup modules
import unittest
import uuid

import pandas as pd
from pathlib import Path

# Import modules for testing
import finbourne_sdk_utils.lpt.qry_scopes as scopes
import finbourne_sdk_utils.lpt.qry_aggregate_holdings as agg
import finbourne_sdk_utils.lpt.qry_portfolios as portfolios
import finbourne_sdk_utils.lpt.get_instruments as get_instruments
import finbourne_sdk_utils.lpt.qry_portfolio_groups as list_portfolio_group
import finbourne_sdk_utils.lpt.qry_portfolio_properties as get_port_props
import finbourne_sdk_utils.lpt.qry_properties as get_props
import finbourne_sdk_utils.lpt.qry_constituents as get_port_cons
import finbourne_sdk_utils.lpt.qry_holdings as get_port_holdings
import finbourne_sdk_utils.lpt.qry_instr_ids as get_instr_id_types
import finbourne_sdk_utils.lpt.qry_instruments as list_instruments
import finbourne_sdk_utils.lpt.qry_portfolio_commands as get_port_commands
import finbourne_sdk_utils.lpt.qry_target_holdings as get_holdings_adj
import finbourne_sdk_utils.lpt.qry_transactions as get_transactions
import finbourne_sdk_utils.lpt.search_instruments as search_instruments
import finbourne_sdk_utils.lpt.create_group_portfolios as cgp
import finbourne_sdk_utils.lpt.upload_portfolio as up
import finbourne_sdk_utils.lpt.upload_quotes as uq
import finbourne_sdk_utils.lpt.txn_config as txn
import os


class CommandsTests(unittest.TestCase):
    test_scope = "lpt-test-scope"
    test_portfolio = "test_port"
    test_date = "2019-10-10"

    work_dir = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.secrets_file = Path(__file__).parent.parent.parent.joinpath("secrets.json")

        cls.work_dir = os.getcwd()
        cls.test_data_path = Path(__file__).parent
        os.chdir(cls.test_data_path)

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.work_dir)

    def display_df(self, df, decimals=2):
        fmt = "{:,." + str(decimals) + "f}"
        pd.options.display.float_format = fmt.format
        pd.set_option("max_colwidth", None)

        try:
            if len(df) == 1 and len(df.columns) > 5:
                df = df.T
            with pd.option_context("display.width", None, "display.max_rows", 1000):
                return df.fillna("")
        except:
            return df

    def validate_results_df(self, df):
        pass
        # self.assertIsInstance(df, pd.DataFrame)

    def validate_results_dict(self, response):
        self.assertIsInstance(response, dict)
        # self.assertTrue(len(df) > 1)

    def validate_success_string(self, response):
        self.assertTrue(response, "Done!")

    @unittest.skip("Calling list scopes in CI will be a very lengthy test.")
    def test_scopes(self):

        result = scopes.main(
            parse=lambda: scopes.parse(args=["--secrets", f"{self.secrets_file}"]),
            display_df=self.display_df,
        )

        self.validate_results_df(result)

    def test_agg(self):

        result = agg.main(
            parse=lambda: agg.parse(
                args=[
                    self.test_scope,
                    self.test_portfolio,
                    self.test_date,
                    "--secrets",
                    f"{self.secrets_file}",
                ]
            ),
            display_df=self.display_df,
        )

        self.validate_results_df(result)

    def test_portfolios(self):

        result = portfolios.main(
            parse=lambda: portfolios.parse(
                args=[self.test_scope, "--secrets", f"{self.secrets_file}"]
            ),
            display_df=self.display_df,
        )
        self.validate_results_df(result)

    def test_instr(self):

        result = get_instruments.main(
            parse=lambda: get_instruments.parse(
                args=[self.test_scope, "--secrets", f"{self.secrets_file}"]
            ),
            display_df=self.display_df,
        )
        self.validate_results_df(result)

    def test_list_portfolio_group(self):

        result = list_portfolio_group.main(
            parse=lambda: list_portfolio_group.parse(
                args=[self.test_scope, "--secrets", f"{self.secrets_file}"]
            ),
            display_df=self.display_df,
        )
        self.validate_results_df(result)

    def test_get_port_props(self):

        result = get_port_props.main(
            parse=lambda: get_port_props.parse(
                args=[
                    self.test_scope,
                    self.test_portfolio,
                    "--secrets",
                    f"{self.secrets_file}",
                ]
            ),
            display_df=self.display_df,
        )
        self.validate_results_df(result)

    @unittest.skip(
        "Jira: SENG-40 - Property search was deprecated and this query requires updating"
    )
    def test_get_props(self):

        result = get_props.main(
            parse=lambda: get_props.parse(args=["--secrets", f"{self.secrets_file}"]),
            display_df=self.display_df,
        )
        self.validate_results_df(result)

    def test_get_port_cons(self):

        result = get_port_cons.main(
            parse=lambda: get_port_cons.parse(
                args=[
                    self.test_scope,
                    self.test_portfolio,
                    self.test_date,
                    "--secrets",
                    f"{self.secrets_file}",
                ]
            ),
            display_df=self.display_df,
        )
        self.validate_results_df(result)

    def test_get_port_holdings(self):

        result = get_port_holdings.main(
            parse=lambda: get_port_holdings.parse(
                args=[
                    self.test_scope,
                    self.test_portfolio,
                    "--secrets-file",
                    f"{self.secrets_file}",
                ]
            ),
            display_df=self.display_df,
        )
        self.validate_results_df(result)

    def test_get_instr_id_types(self):

        result = get_instr_id_types.main(
            parse=lambda: get_instr_id_types.parse(
                args=["--secrets", f"{self.secrets_file}",]
            ),
            display_df=self.display_df,
        )
        self.validate_results_df(result)

    @unittest.skip("Listing all instruments in CI may be a very lengthy process")
    def test_list_instruments(self):

        result = list_instruments.main(
            parse=lambda: list_instruments.parse(
                args=["--secrets", f"{self.secrets_file}",]
            ),
            display_df=self.display_df,
        )
        self.validate_results_df(result)

    def test_get_port_commands(self):

        result = get_port_commands.main(
            parse=lambda: get_port_commands.parse(
                args=[self.test_scope, "abc", "--secrets", f"{self.secrets_file}"]
            ),
            display_df=self.display_df,
        )
        self.validate_results_dict(result)

    def test_get_holdings_adj(self):

        result = get_holdings_adj.main(
            parse=lambda: get_holdings_adj.parse(
                args=[
                    self.test_scope,
                    self.test_portfolio,
                    "--secrets",
                    f"{self.secrets_file}",
                ]
            ),
            display_df=self.display_df,
        )

        self.validate_results_df(result)

    def test_get_transactions(self):

        result = get_transactions.main(
            parse=lambda: get_transactions.parse(
                args=[
                    self.test_scope,
                    self.test_portfolio,
                    "--secrets",
                    f"{self.secrets_file}",
                ]
            ),
            display_df=self.display_df,
        )

        self.validate_results_df(result)

    def test_search_instruments(self):

        result = search_instruments.main(
            parse=lambda: search_instruments.parse(
                args=[
                    "--properties",
                    "Instrument/system/Name=test",
                    "--secrets",
                    f"{self.secrets_file}",
                ]
            ),
            display_df=self.display_df,
        )

        self.validate_results_df(result)

    def test_create_group_portfolios(self):

        data_file = (
            Path(__file__).parent.joinpath("data").joinpath("ExampleData.xlsx:Groups")
        )

        cgp.main(
            parse=lambda: cgp.parse(
                args=[
                    self.test_scope,
                    f"{data_file}",
                    "--secrets",
                    f"{self.secrets_file}",
                ]
            ),
            display_df=self.display_df,
        )

    def test_create_portfolio(self):

        portfolio = f"pf-{str(uuid.uuid4())}"

        up.main(
            parse=lambda: up.parse(
                args=[
                    self.test_scope,
                    portfolio,
                    "-c",
                    portfolio,
                    "GBP",
                    "2018-01-01",
                    "--secrets",
                    f"{self.secrets_file}",
                ]
            ),
            display_df=self.display_df,
        )

    def test_upload_quotes(self):

        data_file = (
            Path(__file__).parent.joinpath("data").joinpath("ExampleData.xlsx:Quotes")
        )

        uq.main(
            parse=lambda: uq.parse(
                args=[
                    self.test_scope,
                    f"{data_file}",
                    "--secrets",
                    f"{self.secrets_file}",
                ]
            ),
            display_df=self.display_df,
        )

    def test_round_trip_transaction_config(self) -> None:
        """
        Tests that you can get and then set a transaction type config in a round trip

        :return:
        """

        data_dir = f"{self.test_data_path.joinpath('data')}"
        os.makedirs(data_dir, exist_ok=True)
        file_path = (
            f"{self.test_data_path.joinpath('data').joinpath('transaction_types.yml')}"
        )

        os.chdir(data_dir)

        # these files get created if something goes wrong in the round trip from change > write > load
        orig_yaml = f"{self.test_data_path.joinpath('orig')}"
        copy_yaml = f"{self.test_data_path.joinpath('copy')}"

        # Ensure that file does not exist
        self.assertFalse(os.path.exists(file_path), msg=f"{file_path} already exists")
        self.assertFalse(
            os.path.exists(orig_yaml),
            msg=f"{orig_yaml} created, implied problem generating yaml",
        )
        self.assertFalse(
            os.path.exists(copy_yaml),
            msg=f"{copy_yaml} created, implied problem generating yaml",
        )

        # Get the transaction types
        txn.main(
            parse=lambda: txn.parse(
                args=["get", "-f", file_path, "--secrets", f"{self.secrets_file}"]
            )
        )

        # Ensure that the file has been created
        self.assertTrue(expr=os.path.exists(file_path))

        # Set the transaction types
        txn.main(
            parse=lambda: txn.parse(
                args=["set", "-f", file_path, "--secrets", f"{self.secrets_file}"]
            )
        )

        # Remove the file and ensure that it is gone
        os.remove(file_path)
        self.assertFalse(expr=os.path.exists(file_path))


if __name__ == "__main__":
    unittest.main()



import unittest


class TestTopLevelImports(unittest.TestCase):
    """Test that the top-level package imports all subpackages."""

    def test_import_finbourne_sdk_utils(self):
        import finbourne_sdk_utils
        self.assertTrue(hasattr(finbourne_sdk_utils, "cocoon"))

    def test_import_cocoon_subpackage(self):
        import finbourne_sdk_utils.cocoon
        self.assertIsNotNone(finbourne_sdk_utils.cocoon)

    def test_import_extract_subpackage(self):
        import finbourne_sdk_utils.extract
        self.assertIsNotNone(finbourne_sdk_utils.extract)

    def test_import_iam_subpackage(self):
        import finbourne_sdk_utils.iam
        self.assertIsNotNone(finbourne_sdk_utils.iam)

    def test_import_jupyter_tools_subpackage(self):
        import finbourne_sdk_utils.jupyter_tools
        self.assertIsNotNone(finbourne_sdk_utils.jupyter_tools)

    def test_import_logger_subpackage(self):
        import finbourne_sdk_utils.logger
        self.assertIsNotNone(finbourne_sdk_utils.logger)

    def test_import_lpt_subpackage(self):
        import finbourne_sdk_utils.lpt
        self.assertIsNotNone(finbourne_sdk_utils.lpt)

    def test_import_pandas_utils_subpackage(self):
        import finbourne_sdk_utils.pandas_utils
        self.assertIsNotNone(finbourne_sdk_utils.pandas_utils)


class TestCocoonInitImports(unittest.TestCase):
    """Test that cocoon/__init__.py exports all expected names."""

    def test_import_load_from_data_frame(self):
        from finbourne_sdk_utils.cocoon import load_from_data_frame
        self.assertTrue(callable(load_from_data_frame))

    def test_import_resolve_instruments(self):
        from finbourne_sdk_utils.cocoon import resolve_instruments
        self.assertTrue(callable(resolve_instruments))

    def test_import_create_property_values(self):
        from finbourne_sdk_utils.cocoon import create_property_values
        self.assertTrue(callable(create_property_values))

    def test_import_set_attributes_recursive(self):
        from finbourne_sdk_utils.cocoon import set_attributes_recursive
        self.assertTrue(callable(set_attributes_recursive))

    def test_import_checkargs(self):
        from finbourne_sdk_utils.cocoon import checkargs
        self.assertTrue(callable(checkargs))

    def test_import_load_data_to_df_and_detect_delimiter(self):
        from finbourne_sdk_utils.cocoon import load_data_to_df_and_detect_delimiter
        self.assertTrue(callable(load_data_to_df_and_detect_delimiter))

    def test_import_check_mapping_fields_exist(self):
        from finbourne_sdk_utils.cocoon import check_mapping_fields_exist
        self.assertTrue(callable(check_mapping_fields_exist))

    def test_import_parse_args(self):
        from finbourne_sdk_utils.cocoon import parse_args
        self.assertTrue(callable(parse_args))

    def test_import_identify_cash_items(self):
        from finbourne_sdk_utils.cocoon import identify_cash_items
        self.assertTrue(callable(identify_cash_items))

    def test_import_validate_mapping_file_structure(self):
        from finbourne_sdk_utils.cocoon import validate_mapping_file_structure
        self.assertTrue(callable(validate_mapping_file_structure))

    def test_import_get_delimiter(self):
        from finbourne_sdk_utils.cocoon import get_delimiter
        self.assertTrue(callable(get_delimiter))

    def test_import_scale_quote_of_type(self):
        from finbourne_sdk_utils.cocoon import scale_quote_of_type
        self.assertTrue(callable(scale_quote_of_type))

    def test_import_strip_whitespace(self):
        from finbourne_sdk_utils.cocoon import strip_whitespace
        self.assertTrue(callable(strip_whitespace))

    def test_import_load_json_file(self):
        from finbourne_sdk_utils.cocoon import load_json_file
        self.assertTrue(callable(load_json_file))

    def test_import_default_fx_forward_model(self):
        from finbourne_sdk_utils.cocoon import default_fx_forward_model
        self.assertTrue(callable(default_fx_forward_model))

    def test_import_format_holdings_response(self):
        from finbourne_sdk_utils.cocoon import format_holdings_response
        self.assertTrue(callable(format_holdings_response))

    def test_import_format_instruments_response(self):
        from finbourne_sdk_utils.cocoon import format_instruments_response
        self.assertTrue(callable(format_instruments_response))

    def test_import_format_portfolios_response(self):
        from finbourne_sdk_utils.cocoon import format_portfolios_response
        self.assertTrue(callable(format_portfolios_response))

    def test_import_format_quotes_response(self):
        from finbourne_sdk_utils.cocoon import format_quotes_response
        self.assertTrue(callable(format_quotes_response))

    def test_import_format_transactions_response(self):
        from finbourne_sdk_utils.cocoon import format_transactions_response
        self.assertTrue(callable(format_transactions_response))

    def test_import_seed_data(self):
        from finbourne_sdk_utils.cocoon import seed_data
        self.assertTrue(callable(seed_data))

    def test_import_cocoon_module(self):
        from finbourne_sdk_utils.cocoon import cocoon
        self.assertIsNotNone(cocoon)

    def test_import_instruments_module(self):
        from finbourne_sdk_utils.cocoon import instruments
        self.assertIsNotNone(instruments)

    def test_import_properties_module(self):
        from finbourne_sdk_utils.cocoon import properties
        self.assertIsNotNone(properties)

    def test_import_systemConfiguration_module(self):
        from finbourne_sdk_utils.cocoon import systemConfiguration
        self.assertIsNotNone(systemConfiguration)

    def test_import_utilities_module(self):
        from finbourne_sdk_utils.cocoon import utilities
        self.assertIsNotNone(utilities)

    def test_import_async_tools_module(self):
        from finbourne_sdk_utils.cocoon import async_tools
        self.assertIsNotNone(async_tools)

    def test_import_validator_module(self):
        from finbourne_sdk_utils.cocoon import validator
        self.assertIsNotNone(validator)

    def test_import_dateorcutlabel_module(self):
        from finbourne_sdk_utils.cocoon import dateorcutlabel
        self.assertIsNotNone(dateorcutlabel)

    def test_import_transaction_type_upload_module(self):
        from finbourne_sdk_utils.cocoon import transaction_type_upload
        self.assertIsNotNone(transaction_type_upload)

    def test_import_create_transaction_type_configuration(self):
        from finbourne_sdk_utils.cocoon.transaction_type_upload import create_transaction_type_configuration
        self.assertTrue(callable(create_transaction_type_configuration))

    def test_import_upsert_transaction_type_alias(self):
        from finbourne_sdk_utils.cocoon.transaction_type_upload import upsert_transaction_type_alias
        self.assertTrue(callable(upsert_transaction_type_alias))


class TestExtractInitImports(unittest.TestCase):
    """Test that extract/__init__.py exports expected names."""

    def test_import_get_holdings_for_group(self):
        from finbourne_sdk_utils.extract import get_holdings_for_group
        self.assertTrue(callable(get_holdings_for_group))


class TestIamInitImports(unittest.TestCase):
    """Test that iam/__init__.py exports expected names."""

    def test_import_roles_module(self):
        from finbourne_sdk_utils.iam import roles
        self.assertIsNotNone(roles)

    def test_import_create_role(self):
        from finbourne_sdk_utils.iam.roles import create_role
        self.assertTrue(callable(create_role))


class TestJupyterToolsInitImports(unittest.TestCase):
    """Test that jupyter_tools/__init__.py exports expected names."""

    def test_import_StopExecution(self):
        from finbourne_sdk_utils.jupyter_tools import StopExecution
        self.assertTrue(isinstance(StopExecution, type))

    def test_import_toggle_code(self):
        from finbourne_sdk_utils.jupyter_tools import toggle_code
        self.assertTrue(callable(toggle_code))


class TestLoggerInitImports(unittest.TestCase):
    """Test that logger/__init__.py exports expected names."""

    def test_import_LusidLogger(self):
        from finbourne_sdk_utils.logger import LusidLogger
        self.assertTrue(isinstance(LusidLogger, type))


class TestLptInitImports(unittest.TestCase):
    """Test that lpt/__init__.py exports expected module attributes."""

    def test_import_lpt_module(self):
        from finbourne_sdk_utils.lpt import lpt
        self.assertIsNotNone(lpt)

    def test_import_either_module(self):
        from finbourne_sdk_utils.lpt import either
        self.assertIsNotNone(either)

    def test_import_stdargs_module(self):
        from finbourne_sdk_utils.lpt import stdargs
        self.assertIsNotNone(stdargs)

    def test_import_record_module(self):
        from finbourne_sdk_utils.lpt import record
        self.assertIsNotNone(record)

    def test_import_pager_module(self):
        from finbourne_sdk_utils.lpt import pager
        self.assertIsNotNone(pager)

    def test_import_refreshing_token_module(self):
        from finbourne_sdk_utils.lpt import refreshing_token
        self.assertIsNotNone(refreshing_token)

    def test_import_connect_lusid_module(self):
        from finbourne_sdk_utils.lpt import connect_lusid
        self.assertIsNotNone(connect_lusid)

    def test_import_dfq_module(self):
        from finbourne_sdk_utils.lpt import dfq
        self.assertIsNotNone(dfq)

    def test_import_lse_module(self):
        from finbourne_sdk_utils.lpt import lse
        self.assertIsNotNone(lse)


class TestPandasUtilsInitImports(unittest.TestCase):
    """Test that pandas_utils/__init__.py exports expected names."""

    def test_import_lusid_response_to_data_frame(self):
        from finbourne_sdk_utils.pandas_utils import lusid_response_to_data_frame
        self.assertTrue(callable(lusid_response_to_data_frame))


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from pathlib import Path
import pandas as pd

from finbourne_sdk_utils import cocoon as cocoon
from parameterized import parameterized
import lusid
import lusid.models as models

from finbourne_sdk_utils import logger
from datetime import datetime
import pytz


class ComplexInstrumentTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.api_factory = lusid.SyncApiClientFactory()
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))
        cls.instruments_api = cls.api_factory.build(lusid.api.InstrumentsApi)

        # Set test parameters as class methods
        cls.scope = "TestScope"
        cls.mapping_required = {
            "name": "Name",
            "definition.start_date": "Issue Date",
            "definition.maturity_date": "Maturity Date",
            "definition.dom_ccy": "Currency",
            "definition.instrument_type": "Instrument Type",
            "definition.principal": "Principal",
            "definition.coupon_rate": "Coupon",
            "definition.identifiers.ClientInternal": "Client Internal",
            "definition.flow_conventions.currency": "Currency",
            "definition.flow_conventions.payment_frequency": "Payment Frequency",
            "definition.flow_conventions.day_count_convention": "Day Count Convention",
            "definition.flow_conventions.roll_convention": "Roll Convention",
            "definition.flow_conventions.payment_calendars": "Payment Calendars",
            "definition.flow_conventions.reset_calendars": "Reset Calendars",
            "definition.flow_conventions.settle_days": "Settlement Days",
            "definition.flow_conventions.reset_days": "Reset Days",
        }
        cls.mapping_optional = {}
        cls.identifier_mapping = {"Isin": "ISIN", "ClientInternal": "Client Internal"}
        cls.file_name = "data/global-fund-fixed-income-master.csv"

    # Set the expected result object using LUSID sdk
    def expected_response(self):
        instrument_definition = models.Bond(
            start_date=datetime(2002, 2, 28, tzinfo=pytz.utc),
            maturity_date=datetime(2032, 3, 1, tzinfo=pytz.utc),
            dom_ccy="USD",
            coupon_rate=0.07,
            principal=500000000,
            flow_conventions=models.FlowConventions(
                # coupon payment currency
                currency="USD",
                # semi-annual coupon payments
                payment_frequency="6M",
                # using an Actual/365 day count convention (other options : Act360, ActAct, ...
                day_count_convention="Thirty360",
                # modified following rolling convention (other options : ModifiedPrevious, NoAdjustment, EndOfMonth,...)
                roll_convention="MF",
                # no holiday calendar supplied
                payment_calendars=["US"],
                reset_calendars=["US"],
                settle_days=2,
                reset_days=2,
            ),
            identifiers={"clientInternal": "imd_34534538"},
            instrument_type="Bond",
        )

        instrument_request = {
            "ClientInternal: imd_34534538": models.InstrumentDefinition(
                # instrument display name
                name="DISNEY_7_2032",
                # unique instrument identifier
                identifiers={
                    "ClientInternal": models.InstrumentIdValue(value="imd_34534538"),
                    "Isin": models.InstrumentIdValue(value="US25468PBW59"),
                },
                # instrument definition
                definition=instrument_definition,
            )
        }

        return self.instruments_api.upsert_instruments(instrument_request)

    @unittest.skip("Also failed on the original V1, fixup later")
    def test_load_from_dataframe_bond_v1_fail(self) -> None:
        """
            Test that instruments can be loaded successfully using load_from_data_frame()

            :param str scope: The scope of the portfolios to load the transactions into
            :param str file_name: The name of the test data file
            :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
            :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
            :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
            :param list[str] property_columns: The columns to create properties for
            :param str properties_scope: The scope to add the properties to

            :return: None
            """
        # Load parameters
        scope = self.scope
        mapping_required = self.mapping_required
        mapping_optional = self.mapping_optional
        identifier_mapping = self.identifier_mapping
        file_name = self.file_name
        # Load expected outcome for an instrument
        expected_outcome = self.expected_response()

        # Load data frame with instrument data
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="instruments",
            identifier_mapping=identifier_mapping,
            instrument_name_enrichment=True,
        )

        self.assertEqual(
            first=sum(
                [
                    len(response.values)
                    for response in responses["instruments"]["success"]
                ]
            ),
            second=len(data_frame["Client Internal"].unique()),
        )

        self.assertEqual(
            first=sum(
                [
                    len(response.failed)
                    for response in responses["instruments"]["success"]
                ]
            ),
            second=0,
        )

        # Clean and aggregate successful responses into a dictionary
        combined_successes = {
            correlation_id: instrument
            for response in responses["instruments"]["success"]
            for correlation_id, instrument in response.values.items()
        }

        # Check instrument definition in response matches expected outcome
        for key in expected_outcome.values.keys():
            response_obj = combined_successes[key]
            self.assertEqual(
                response_obj.instrument_definition,
                expected_outcome.values[key].instrument_definition,
            )

    def test_load_from_dataframe_with_unsupported_inst_type(self) -> None:
        """
            Tests that an invalid instrument_type raises a ValueError in load_from_data_frame()

            :param str scope: The scope of the portfolios to load the transactions into
            :param str file_name: The name of the test data file
            :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
            :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
            :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
            :param list[str] property_columns: The columns to create properties for
            :param str properties_scope: The scope to add the properties to

            :return: None
            """

        # Load parameters
        scope = self.scope
        mapping_required = self.mapping_required
        mapping_optional = self.mapping_optional
        identifier_mapping = self.identifier_mapping
        file_name = self.file_name

        # Load data frame with instrument data
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))
        # Overwrite "Instrument Type" to an unsupported value
        data_frame["Instrument Type"].replace("Bond", "Foo", inplace=True)

        # Check that ValueError is raised
        with self.assertRaises(ValueError):
            cocoon.cocoon.load_from_data_frame(
                api_factory=self.api_factory,
                scope=scope,
                data_frame=data_frame,
                mapping_required=mapping_required,
                mapping_optional=mapping_optional,
                file_type="instruments",
                identifier_mapping=identifier_mapping,
                instrument_name_enrichment=True,
            )


if __name__ == "__main__":
    unittest.main()



import os

import pandas as pd

from finbourne_sdk_utils.lpt import lpt
from finbourne_sdk_utils.lpt import lse
from finbourne_sdk_utils.lpt import stdargs
from finbourne_sdk_utils.lpt.either import Either

mapping_prefixes = {"Figi": "FIGI", "ClientInternal": "INT", "QuotePermId": "QPI"}

mapping_table = {}


def parse(extend=None, args=None):
    return (
        stdargs.Parser("Map Instruments", ["filename"])
        .add("--folder", help="include all 'txn' files in the folder")
        .add("input", nargs="*", metavar="file", help="file(s) containing instruments")
        .add(
            "--column",
            metavar="input-column",
            default="instrument_uid",
            help="column name for instrument column",
        )
        .extend(extend)
        .parse(args)
    )


def process_args(api, args):
    if args.folder:
        args.input.extend(
            [
                os.path.join(args.folder, f)
                for f in os.listdir(args.folder)
                if "-txn-" in f and f.endswith(".csv")
            ]
        )

    df = (
        pd.concat(
            [lpt.read_csv(f)[[args.column]].drop_duplicates() for f in args.input],
            ignore_index=True,
            sort=True,
        )
        .drop_duplicates()
        .reset_index(drop=True)
    )

    df.columns = ["FROM"]
    df["TO"] = df["FROM"]

    return map_instruments(api, df, "TO")


def main():
    lpt.standard_flow(parse, lse.connect, process_args)


def map_instruments(api, df, column):
    WORKING = "__:::__"  # temporary column name

    # Apply any known mappings to avoid unecessary i/o

    if len(mapping_table) > 0:
        srs = df[column].map(mapping_table)
        srs = srs[srs.notnull()]
        if len(srs) > 0:
            df.loc[srs.index, column] = srs

    # updates the mappings table
    def update_mappings(src, prefix):
        mapping_table.update(
            {prefix + k: v.lusid_instrument_id for k, v in src.items()}
        )

    def batch_query(instr_type, prefix, outstanding):
        if len(outstanding) > 0:
            batch = outstanding[:500]  # records to process now
            remainder = outstanding[500:]  # remaining records

            # called if the get_instruments() succeeds
            def get_success(result):
                get_found = result.content.values
                get_failed = result.content.failed

                # Update successfully found items
                update_mappings(get_found, prefix)

                if len(get_failed) > 0:
                    if instr_type == "ClientInternal":
                        # For un-mapped internal codes, we will try to add (upsert)

                        # called if the upsert_instruments() succeeds
                        def add_success(result):
                            add_worked = result.content.values
                            add_failed = result.content.failed

                            if len(add_failed) > 0:
                                return Either.Left("Failed to add internal instruments")

                            # Update successfully added items
                            update_mappings(add_worked, prefix)

                            # Kick off the next batch
                            return batch_query(instr_type, prefix, remainder)

                        # Create the upsert request from the failed items
                        request = {
                            k: api.models.InstrumentDefinition(
                                name=v.id, identifiers={"ClientInternal": v.id}
                            )
                            for k, v in get_failed.items()
                        }

                        return api.call.upsert_instruments(request).bind(add_success)
                    else:
                        # Instruments are not mapped. Nothing we cando.
                        return Either.Left(
                            "Failed to locate instruments of type {}".format(instr_type)
                        )
                else:
                    # No failures, kick off the next batch
                    return batch_query(instr_type, prefix, remainder)

            return api.call.get_instruments(
                instr_type, list(batch[WORKING].values)
            ).bind(get_success)
        else:
            # No records remaining. Return the now-enriched dataframe
            return Either.Right(df)

    def map_type(key, instr_type):
        prefix = key + ":"
        subset = df[df[column].str.startswith(prefix)]

        # See if there are any entries of this type
        if len(subset) > 0:
            width = len(prefix)
            uniques = subset[[column]].drop_duplicates(column)
            uniques[WORKING] = uniques[column].str.slice(width)

            def map_success(v):
                df.loc[subset.index, column] = subset[column].map(mapping_table)
                return Either.Right(df)

            return batch_query(instr_type, prefix, uniques).bind(map_success)
        else:
            # Nothing to be done, pass the full result back
            return Either.Right(df)

    return (
        map_type("FIGI", "Figi")
        .bind(lambda r: map_type("INT", "ClientInternal"))
        .bind(lambda r: map_type("QPI", "QuotePermId"))
    )


def include_mappings(path):
    if path:
        df = lpt.read_csv(path)

        mapping_table.update(df.set_index("FROM")["TO"].to_dict())



from http import HTTPStatus

import lusid


class MockApiFactory(lusid.SyncApiClientFactory):
    """
    This is a mock of the lusid.SyncApiClientFactory class
    """

    def build(self, api):
        """
        Creates and returns appropriate Mock for api passed in
        supports:
         - lusid.PropertyDefinitionsApi
        :param lusid.api api: The api to mock

        :return: mock(lusid.api): The mocked api
        """

        if api == lusid.PropertyDefinitionsApi:
            return self.MockPropertyDefinitionsApi()

    class MockPropertyDefinitionsApi:
        """
        A mock of the lusid.PropertyDefinitionsApi
        """

        def create_property_definition(
            self, create_property_definition_request
        ) -> lusid.models.PropertyDefinition:
            """
            This mocks the creation of a portfolio definition

            :param lusid.models.CreatePropertyDefinitionRequest create_property_definition_request: The create property definition request
            :param kwargs:
            :return: lusid.models.PropertyDefinition: The property defintion of the created property
            """
            return lusid.models.PropertyDefinition(
                key=f"{create_property_definition_request.domain}/{create_property_definition_request.scope}/{create_property_definition_request.code}",
                data_type_id=lusid.models.ResourceId(
                    scope=create_property_definition_request.data_type_id.scope,
                    code=create_property_definition_request.data_type_id.code,
                ),
            )

        def get_property_definition(
            self, domain, scope, code
        ) -> lusid.models.PropertyDefinition:
            """
            This mocks the call to get a property definition

            :param domain: The domain of the property
            :param scope: The scope of the property
            :param code: The code of the property
            :param kwargs:

            :return: lusid.models.PropertyDefinition any: The property definition of the property if it exists
            """
            # Construct the property key
            property_key = f"{domain}/{scope}/{code}"

            # A static representation of the property definitions that exist
            property_keys_in_existance = {
                "Instrument/default/Figi": lusid.models.ResourceId(
                    scope="system", code="string"
                ),
                "Transaction/default/TradeToPortfolioRate": lusid.models.ResourceId(
                    scope="system", code="number"
                ),
                "Transaction/Operations/Strategy": lusid.models.ResourceId(
                    scope="system", code="string"
                ),
                "Holding/Operations/Currency": lusid.models.ResourceId(
                    scope="system", code="currency"
                ),
            }

            # If the property exists return the defintion, else raise an exception
            if property_key in list(property_keys_in_existance.keys()):
                return lusid.models.PropertyDefinition(
                    key=property_key,
                    data_type_id=property_keys_in_existance[property_key],
                )
            else:
                raise lusid.exceptions.ApiException(status=HTTPStatus.NOT_FOUND)



import os
import unittest
from pathlib import Path
import pandas as pd

from finbourne_sdk_utils import cocoon as cocoon
from parameterized import parameterized
import lusid
from finbourne_sdk_utils import logger
import numpy as np
import concurrent.futures


class CocoonTestsAsyncTools(unittest.TestCase):
    """"
    These tests are to ensure that the Asynchronous Tools work as expected.
    """

    @classmethod
    def setUpClass(cls) -> None:
        secrets_file = Path(__file__).parent.parent.parent.joinpath("secrets.json")
        cls.api_factory = lusid.SyncApiClientFactory( )
        
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    @parameterized.expand(
        [
            [
                "Standard load with ~700 portfolios in 1 batch with max_threads per batch of 5",
                "data/holdings-example-large.csv",
                1,
                5,
            ],
            [
                "Standard load with ~700 portfolios in 5 batches with max_threads per batch of 5",
                "data/holdings-example-large.csv",
                5,
                5,
            ],
            [
                "Standard load with ~700 portfolios in 5 batches with max_threads per batch of 10",
                "data/holdings-example-large.csv",
                5,
                10,
            ],
        ]
    )
    @unittest.skip
    def test_multiple_threads(
        self, _, file_name, number_threads, thread_pool_max_workers
    ):
        """
        This tests different combinations of running load_from_data_frame across multiple threads and configuring
        the max number of workers that each call to load_from_data_frame will use in its thread pool.

        :param str _: The name of the test
        :param str file_name: The name of the test data file working with holdings
        :param int number_threads: The number of threads to split the file load across (the number of times load_from_data_frame is called)
        :param int thread_pool_max_workers: The maximum number of workers per thread pool for each call to load_from_data_frame

        :return: None
        """

        mapping_required = {
            "code": "FundCode",
            "effective_at": "Effective Date",
            "tax_lots.units": "Quantity",
        }

        mapping_optional = {
            "tax_lots.cost.amount": None,
            "tax_lots.cost.currency": "Local Currency Code",
            "tax_lots.portfolio_cost": None,
            "tax_lots.price": None,
            "tax_lots.purchase_date": None,
            "tax_lots.settlement_date": None,
        }

        identifier_mapping = {
            "Isin": "ISIN Security Identifier",
            "Sedol": "SEDOL Security Identifier",
            "Currency": "is_cash_with_currency",
        }

        property_columns = ["Prime Broker"]

        properties_scope = "operations001"

        scope = "prime_broker_test"

        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        # Split the dataframe across the multiple threads
        splitted_arrays = np.array_split(data_frame, number_threads)

        # Create a ThreadPool to run the threads in
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=number_threads)

        # Run each call to load_from_data_frame in a different thread
        futures = [
            executor.submit(
                cocoon.cocoon.load_from_data_frame,
                self.api_factory,
                scope,
                splitted_arrays[i],
                mapping_required,
                mapping_optional,
                "holdings",
                identifier_mapping,
                property_columns,
                properties_scope,
                None,
                True,
                False,
                None,
                False,
                thread_pool_max_workers,
            )
            for i in range(number_threads)
        ]

        # Get the results
        responses = [future.result() for future in futures]

        # Check that the results are as expected
        self.assertGreater(
            sum([len(response["holdings"]["success"]) for response in responses]), 0
        )
        self.assertEqual(
            sum([len(response["holdings"]["success"]) for response in responses]),
            len(data_frame),
        )
        self.assertEqual(
            sum([len(response["holdings"]["errors"]) for response in responses]), 0
        )

        self.assertTrue(
            expr=all(
                isinstance(succcess_response.version, lusid.models.Version)
                for response in responses
                for succcess_response in response["holdings"]["success"]
            )
        )



import os
import unittest
import json
from pathlib import Path
import pandas as pd
from datetime import datetime
import pytz

from finbourne_sdk_utils import cocoon as cocoon
from parameterized import parameterized
import lusid
from finbourne_sdk_utils import logger
from finbourne_sdk_utils.cocoon.utilities import create_scope_id


def holding_instrument_identifiers(fake1=False, fake2=False):
    if fake1:
        fake1 = {
            "Instrument/default/Sedol": "FAKESEDOL1",
            "Instrument/default/Isin": "FAKEISIN1",
        }
    else:
        fake1 = None

    if fake2:
        fake2 = {
            "Instrument/default/Sedol": "FAKESEDOL2",
            "Instrument/default/Isin": "FAKEISIN2",
        }
    else:
        fake2 = None

    return [instrument for instrument in [fake1, fake2] if instrument]


def extract_unique_identifiers_from_holdings_response(response):
    unmatched_instruments = [
        holding_adjustment.instrument_identifiers
        for holding_adjustment in response["holdings"].get("unmatched_items", [])
    ]

    return list(
        map(
            dict,
            set(
                tuple(sorted(identifiers.items()))
                for identifiers in unmatched_instruments
            ),
        )
    )


class CocoonTestsHoldings(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        
        cls.api_factory = lusid.SyncApiClientFactory()
        
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    @parameterized.expand(
        [
            [
                "Standard successful load",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Standard load with ~700 portfolios",
                "prime_broker_test",
                "data/holdings-example-large.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Standard successful load with adjustment only",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                True,
                lusid.models.Version,
            ],
            [
                "Standard successful load with string index",
                "prime_broker_test",
                "data/holdings-example-unique-date-string-index.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Standard successful load with duplicate index",
                "prime_broker_test",
                "data/holdings-example-unique-date-duplicate-index.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Standard successful load with unique properties scope",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                f"operations001_{create_scope_id()}",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Add in some constants",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": 3000,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Add in a default value for a required field with no specified column",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": {"default": "2019-10-08"},
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Add in a column as a nested dictionary",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": {"column": "Effective Date"},
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Add in a column and default as a nested dictionary",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": {
                        "column": "Effective Date",
                        "default": "2019-10-08",
                    },
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Multiple effective dates in a holdings file",
                "prime_broker_test",
                "data/holdings-example.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Standard successful load with sub-holding-keys based on a column that exists in the file",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                ["Security Description"],
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Standard successful load with sub-holding-keys based on two columns that exists in the file",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                ["Security Description", "Prime Broker"],
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Standard successful load with a sub-holding-key that is not populated for all rows",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                ["swap"],
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Standard successful load with a sub-holding-key that has no pre-existing property definition",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                f"operations001_{create_scope_id()}",
                ["Prime Broker"],
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Standard successful load with a sub-holding-key that has no pre-existing property definition and a different scope",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                f"operations001",
                ["Prime Broker"],
                f"accountview_{create_scope_id()}",
                False,
                lusid.models.Version,
            ],
            [
                "Standard successful load with a sub-holding-key that has a metric value",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                ["Quantity"],
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Duplicate column in the source file",
                "prime_broker_test",
                "data/holdings-example-unique-date-duplicate-column.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Pass string as tax_lots.units value",
                "prime_broker_test",
                "data/holdings-example-unique-date-duplicate-column.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": 1,
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Pass integer as tax_lots.units value",
                "prime_broker_test",
                "data/holdings-example-unique-date-duplicate-column.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": 1,
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
            [
                "Standard successful load but with one unmatched identifier. No flag for return_unmatched_identifiers"
                "means that we should not get that unmatched_identifier returned.",
                "prime_broker_test",
                "data/holdings-example-unique-date-one-unmatched-instrument.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                None,
                None,
                False,
                lusid.models.Version,
            ],
        ]
    )
    def test_load_from_data_frame_holdings_success(
        self,
        _,
        scope,
        file_name,
        mapping_required,
        mapping_optional,
        identifier_mapping,
        property_columns,
        properties_scope,
        sub_holding_keys,
        sub_holding_key_scope,
        holdings_adjustment_only,
        expected_outcome,
    ) -> None:
        """
        Test that holdings can be loaded successfully

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list[str] property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to
        :param list sub_holding_keys: The sub holding keys to populate on the adjustments as transaction properties
        :param bool holdings_adjustment_only: Whether to use the adjust_holdings api call rather than set_holdings when
               working with holdings
        :param any expected_outcome: The expected outcome

        :return: None
        """
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="holdings",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            sub_holding_keys=sub_holding_keys,
            holdings_adjustment_only=holdings_adjustment_only,
            sub_holding_keys_scope=sub_holding_key_scope,
        )

        self.assertGreater(len(responses["holdings"]["success"]), 0)

        self.assertEqual(
            len(responses["holdings"]["errors"]), 0, responses["holdings"]["errors"]
        )

        # Assert that by default no unmatched_identifiers are returned in the response
        self.assertFalse(responses["holdings"].get("unmatched_identifiers", False))

        self.assertTrue(
            expr=all(
                isinstance(success_response.version, expected_outcome)
                for success_response in responses["holdings"]["success"]
            )
        )

    @parameterized.expand(
        [
            [
                "Standard successful load with no unmatched instruments - SetHolding",
                "data/holdings-example-unique-date.csv",
                None,
                False,
                0,
                holding_instrument_identifiers(fake1=False, fake2=False),
            ],
            [
                "Standard successful load with one unmatched instrument - SetHolding",
                "data/holdings-example-unique-date-one-unmatched-instrument.csv",
                None,
                False,
                1,
                holding_instrument_identifiers(fake1=True, fake2=False),
            ],
            [
                "Standard successful load with one instrument that has a correct Isin and ClientInternal "
                "but fake Sedol so it should still resolve and return no unmatched identifiers - SetHolding",
                "data/holdings-example-unique-date-incorrect-sedol-correct_isin_clientInternal_should_resolve.csv",
                None,
                False,
                0,
                holding_instrument_identifiers(fake1=False, fake2=False),
            ],
            [
                "Standard successful load with one unmatched instrument in two separate adjustments (two portfolios) - SetHolding",
                "data/holdings-example-unique-date-one-unmatched-instrument-duplicated-in-two-adjustments.csv",
                None,
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=False),
            ],
            [
                "Standard successful load with two unmatched instrument in two separate adjustments (two portfolios) - SetHolding",
                "data/holdings-example-unique-date-two-unmatched-instruments-separate-adjustments.csv",
                None,
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=True),
            ],
            [
                "Standard successful load with two unmatched instrument in one single adjustment (one portfolio) - SetHolding",
                "data/holdings-example-unique-date-two-unmatched-instruments-single-adjustment.csv",
                None,
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=True),
            ],
            [
                "Standard successful load with one unmatched instrument in two adjustments across two dates (one portfolio) - SetHolding",
                "data/holdings-example-one-unmatched-instrument-in-two-adjustments-one-portfolio-two-dates.csv",
                None,
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=False),
            ],
            [
                "Standard successful load with one unmatched instrument in two adjustments one date (two portfolios) - SetHolding",
                "data/holdings-example-same-date-one-unmatched-instrument-duplicated-in-two-adjustments-two-portfolios.csv",
                None,
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=False),
            ],
            [
                "Standard successful load with two unmatched instrument in two adjustments two dates (two portfolios) - SetHolding",
                "data/holdings-example-two-dates-two-unmatched-instruments-separate-adjustments-two-portfolios.csv",
                None,
                False,
                4,
                holding_instrument_identifiers(fake1=True, fake2=True),
            ],
            [
                "One portfolio, one sub-holding-key, one date, two instruments - SetHolding",
                "data/holdings-example-one-portfolio-one-shk-one-date-two-instruments.csv",
                ["Security Description"],
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=True),
            ],
            [
                "One portfolio, one sub-holding-key, two dates, two instruments - SetHolding",
                "data/holdings-example-one-portfolio-one-shk-two-dates-two-instruments.csv",
                ["Security Description"],
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=True),
            ],
            [
                "One portfolio, two sub-holding-keys, one date, same instrument - SetHolding",
                "data/holdings-example-one-portfolio-two-shks-one-date-one-instrument.csv",
                ["Security Description"],
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=False),
            ],
            [
                "One portfolio, two sub-holding-keys, one date, two instruments - SetHolding",
                "data/holdings-example-one-portfolio-two-shks-one-date-two-instruments.csv",
                ["Security Description"],
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=True),
            ],
            [
                "One portfolio, two sub-holding-keys, two dates, two instruments - SetHolding",
                "data/holdings-example-one-portfolio-two-shks-two-dates-two-instruments.csv",
                ["Security Description"],
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=True),
            ],
            [
                "One portfolio, two sub-holding-keys, two dates, same instrument - SetHolding",
                "data/holdings-example-one-portfolio-two-shks-two-dates-one-instrument.csv",
                ["Security Description"],
                False,
                2,
                holding_instrument_identifiers(fake1=True, fake2=False),
            ],
            [
                "One portfolio, two sub-holding-keys, two dates, same instrument - AdjustHolding",
                "data/holdings-example-one-portfolio-two-shks-two-dates-one-instrument.csv",
                ["Security Description"],
                True,
                2,
                holding_instrument_identifiers(fake1=True, fake2=False),
            ],
            [
                "One portfolio, one sub-holding-key, two dates, two instruments - AdjustHolding",
                "data/holdings-example-one-portfolio-one-shk-two-dates-two-instruments.csv",
                ["Security Description"],
                True,
                2,
                holding_instrument_identifiers(fake1=True, fake2=True),
            ],
            [
                "Standard successful load with two unmatched instrument in one single adjustment (one portfolio) - AdjustHolding",
                "data/holdings-example-unique-date-two-unmatched-instruments-single-adjustment.csv",
                None,
                True,
                2,
                holding_instrument_identifiers(fake1=True, fake2=True),
            ],
            [
                "Standard successful load with one unmatched instrument - AdjustHolding",
                "data/holdings-example-unique-date-one-unmatched-instrument.csv",
                None,
                True,
                1,
                holding_instrument_identifiers(fake1=True, fake2=False),
            ],
            [
                "Standard successful load with two unmatched instruments across 3 unique identifiers - AdjustHolding",
                "data/holdings-example-unique-date-two-unmatched-instruments-single-adjustment_three_unique_ids.csv",
                None,
                True,
                2,
                [
                    {
                        "Instrument/default/Sedol": "FAKESEDOL1",
                        "Instrument/default/Isin": "FAKEISIN1",
                    },
                    {
                        "Instrument/default/Sedol": "FAKESEDOL1",
                        "Instrument/default/Isin": "FAKEISIN2",
                    },
                ],
            ],
        ]
    )
    def test_load_from_data_frame_holdings_success_with_unmatched_items(
        self,
        _,
        file_name,
        sub_holding_keys,
        holdings_adjustment_only,
        expected_holdings_in_response,
        expected_unmatched_identifiers,
    ) -> None:
        """
        Test that holdings can be loaded successfully

        :param str file_name: The name of the test data file
        :param list[str] sub_holding_keys: The column names for any sub-holding-keys
        :param bool holdings_adjustment_only: Whether to use the set or adjust holding api
        :param int expected_holdings_in_response: The number of expected holdings in the response
        :param list[instrument_identifiers] expected_unmatched_identifiers: A list of expected instrument_identifiers

        :return: None
        """
        # Unchanged vars that have no need to be passed via param (they would count as duplicate lines)
        scope = f"unmatched_holdings_test_{create_scope_id()}"
        mapping_required = {
            "code": "FundCode",
            "effective_at": "Effective Date",
            "tax_lots.units": "Quantity",
        }
        mapping_optional = {"tax_lots.cost.currency": "Local Currency Code"}
        identifier_mapping = {
            "Isin": "ISIN Security Identifier",
            "Sedol": "SEDOL Security Identifier",
            "Currency": "is_cash_with_currency",
            "ClientInternal": "Client Internal",
        }
        property_columns = ["Prime Broker"]
        properties_scope = "operations001"
        sub_holding_key_scope = None
        return_unmatched_items = True
        expected_outcome = lusid.models.Version

        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        # Create instruments that should resolve (for tests expecting instruments to exist)
        # This is specifically for the test with incorrect SEDOL but correct ISIN and ClientInternal
        if "incorrect-sedol-correct_isin_clientInternal_should_resolve" in file_name:
            instruments_df = pd.DataFrame([{
                "name": "12% BD REDEEM 26/07/2014 CRC 1000000",
                "isin": "CRMADAPB2335",
                "clientInternal": "3940007-CRBNV-CRC"
            }])
            
            instrument_response = cocoon.cocoon.load_from_data_frame(
                api_factory=self.api_factory,
                scope=scope,
                data_frame=instruments_df,
                mapping_required={
                    "name": "name"
                },
                mapping_optional={},
                identifier_mapping={
                    "Isin": "isin",
                    "ClientInternal": "clientInternal"
                },
                file_type="instruments"
            )
            
            # Verify instruments were created successfully
            self.assertEqual(len(instrument_response.get("instruments").get("errors")), 0)

        # Create the portfolios
        portfolio_response = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required={
                "code": "FundCode",
                "display_name": "FundCode",
                "base_currency": "Local Currency Code",
            },
            file_type="portfolio",
            mapping_optional={"created": "Created Date"},
        )

        # Assert that all portfolios were created without any issues
        self.assertEqual(len(portfolio_response.get("portfolios").get("errors")), 0)

        # Load in the holdings adjustments
        holding_responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="holdings",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            sub_holding_keys=sub_holding_keys,
            holdings_adjustment_only=holdings_adjustment_only,
            sub_holding_keys_scope=sub_holding_key_scope,
            return_unmatched_items=return_unmatched_items,
        )

        self.assertGreater(len(holding_responses["holdings"]["success"]), 0)

        self.assertEqual(len(holding_responses["holdings"]["errors"]), 0)

        # Assert that the number of holdings returned are as expected
        self.assertEqual(
            len(holding_responses["holdings"]["unmatched_items"]),
            expected_holdings_in_response,
        )

        # Assert that the holdings returned only contain the instrument identifiers that did not resolve
        self.assertCountEqual(
            extract_unique_identifiers_from_holdings_response(holding_responses),
            expected_unmatched_identifiers,
        )

        self.assertTrue(
            expr=all(
                isinstance(success_response.version, expected_outcome)
                for success_response in holding_responses["holdings"]["success"]
            )
        )

        # Delete the portfolios at the end of the test
        for portfolio in portfolio_response.get("portfolios").get("success"):
            self.api_factory.build(lusid.api.PortfoliosApi).delete_portfolio(
                scope=scope, code=portfolio.id.code
            )

    def test_return_unmatched_holding_handles_empty_holdings_exception_in_lusid(self):
        """
        Test that the get holdings call to LUSID handles cases where there are no holdings present for a
        given portfolio code and effective at combination.

        :return: None
        """
        portfolio_code_and_scope = "set_holdings_test"
        effective_at_datetime_for_portfolio = datetime(2020, 1, 1, tzinfo=pytz.utc)
        base_currency = "GBP"
        effective_at_str_for_holdings = "2020-01-01T12:00:00"
        code_tuple = (portfolio_code_and_scope, effective_at_str_for_holdings)

        # Create an empty portfolio for the test, but handle situations where the portfolio already exists
        try:
            portfolio_setup_response = self.api_factory.build(
                lusid.api.TransactionPortfoliosApi
            ).create_portfolio(
                scope=portfolio_code_and_scope,
                create_transaction_portfolio_request=lusid.models.CreateTransactionPortfolioRequest(
                    display_name=portfolio_code_and_scope,
                    description=portfolio_code_and_scope,
                    code=portfolio_code_and_scope,
                    created=effective_at_datetime_for_portfolio,
                    base_currency=base_currency,
                ),
            )

            # Assert that the portfolio was created successfully
            self.assertIsInstance(portfolio_setup_response, lusid.models.Portfolio)

        except lusid.ApiException as e:
            if "PortfolioWithIdAlreadyExists" not in str(e.body):
                raise e

        # Call the return_unmatched_holdings method where there are no holdings
        unmatched_holdings_response = cocoon.cocoon.return_unmatched_holdings(
            api_factory=self.api_factory,
            scope=portfolio_code_and_scope,
            code_tuple=code_tuple,
        )

        # Assert that the unmatched holdings response does not throw an error, and returns an empty list
        self.assertEqual([], unmatched_holdings_response)

        # Delete the portfolio at the end of the test
        self.api_factory.build(lusid.api.PortfoliosApi).delete_portfolio(
            scope=portfolio_code_and_scope, code=portfolio_code_and_scope
        )

    @parameterized.expand(
        [
            [
                "Failed load with duplicate shk",
                "data/holdings-example-duplicate-shk.csv",
                ["Security Description"],
                "DuplicateSubHoldingKeysProvided",
                False,
            ],
            [
                "Failed with non-existent portfolio",
                "data/holdings-example-single-holding.csv",
                ["Security Description"],
                "PortfolioNotFound",
                True,
            ],
        ]
    )
    def test_failed_load_from_data_frame_holdings_returns_useful_errors_and_skips_validation_logic(
        self, _, file_name, sub_holding_keys, error_name, skip_portfolio,
    ) -> None:
        """
        Test that a failed holding upload is handled gracefully by the whole load_from_data_frame function.
        If any upload causes an error the validation logic will not run and all the errors will be returned
        by load_from_data_frame.

        :param str file_name: The name of the test data file
        :param list[str] sub_holding_keys: The column names for any sub-holding-keys
        :param str error_name: The name of the ApiException that is hit by this error
        :param bool skip_portfolio: Flag whether to skip the creation of a test portfolio

        :return: None
        """
        # Unchanged vars that have no need to be passed via param (they would count as duplicate lines)
        scope = f"unmatched_holdings_test_{create_scope_id()}"
        mapping_required = {
            "code": "FundCode",
            "effective_at": "Effective Date",
            "tax_lots.units": "Quantity",
        }
        mapping_optional = {"tax_lots.cost.currency": "Local Currency Code"}
        identifier_mapping = {
            "Isin": "ISIN Security Identifier",
            "Sedol": "SEDOL Security Identifier",
            "Currency": "is_cash_with_currency",
            "ClientInternal": "Client Internal",
        }
        property_columns = ["Prime Broker"]
        properties_scope = "operations001"
        sub_holding_key_scope = None
        return_unmatched_items = True
        holdings_adjustment_only = False
        failed_unmatched_items_check = [
            "Please resolve all upload errors to check for unmatched items."
        ]

        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        if not skip_portfolio:
            # Create the portfolios
            portfolio_response = cocoon.cocoon.load_from_data_frame(
                api_factory=self.api_factory,
                scope=scope,
                data_frame=data_frame,
                mapping_required={
                    "code": "FundCode",
                    "display_name": "FundCode",
                    "base_currency": "Local Currency Code",
                },
                file_type="portfolio",
                mapping_optional={"created": "Created Date"},
            )

            # Assert that all portfolios were created without any issues
            self.assertEqual(len(portfolio_response.get("portfolios").get("errors")), 0)

        # Load in the holdings adjustments
        holding_responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="holdings",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            sub_holding_keys=sub_holding_keys,
            holdings_adjustment_only=holdings_adjustment_only,
            sub_holding_keys_scope=sub_holding_key_scope,
            return_unmatched_items=return_unmatched_items,
        )

        self.assertEqual(len(holding_responses["holdings"]["success"]), 0)

        self.assertEqual(len(holding_responses["holdings"]["errors"]), 1)

        # Assert that the ApiError thrown by LUSID is what is expected, given the input data
        self.assertEqual(
            json.loads(holding_responses["holdings"]["errors"][0].body)["name"],
            error_name,
        )

        # Assert that there is no 'unmatched_item' field if the input data resulted in no successful uploads
        self.assertEqual(
            holding_responses["holdings"].get("unmatched_items"),
            failed_unmatched_items_check,
        )

        if not skip_portfolio:
            # Delete the portfolios at the end of the test
            for portfolio in portfolio_response.get("portfolios").get("success"):
                self.api_factory.build(lusid.api.PortfoliosApi).delete_portfolio(
                    scope=scope, code=portfolio.id.code
                )



import unittest
from finbourne_sdk_utils.jupyter_tools import StopExecution


class TestPortfolioStopper(unittest.TestCase):
    def test_portfolio_stopper(self) -> None:

        portfolio_code = "testPortfolio"
        error_message = f"Portfolio {portfolio_code} does not exist. Stopping notebook"

        with self.assertRaises(StopExecution) as error:
            raise StopExecution(error_message)

        self.assertEqual(error.exception.message, error_message)



import os
import unittest
from pathlib import Path
import pandas as pd
import lusid

from finbourne_sdk_utils import cocoon as cocoon
from parameterized import parameterized
from finbourne_sdk_utils import logger
import asyncio


class CocoonInstrumentsTests(unittest.TestCase):
    api_factory = None

    @classmethod
    def setUpClass(cls) -> None:

        cls.api_factory = lusid.SyncApiClientFactory()
        
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))
        cls.loop = cocoon.async_tools.start_event_loop_new_thread()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.loop.stop()

    def test_resolve_valid_instruments(self):
        source_ids = ["BBG000C6K6G9", "BBG000C04D57"]
        names = ["Inst1", "Inst2"]

        data = {"ids": source_ids, "is_cash": [None, None]}
        data_expected_value = {
            "inst_name": names,
            "ids": source_ids,
            "is_cash": [None, None],
        }
        required_mapping_expected_value = {"name": "inst_name"}

        identifier_mapping_expected_value = {
            "Figi": "ids",
            "Instrument/default/Currency": "is_cash",
        }

        identifier_mapping = {"Figi": "ids", "Instrument/default/Currency": "is_cash"}

        df = pd.DataFrame(data)
        df_expected_value = pd.DataFrame(data_expected_value)
        upsert_response = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            mapping_required=required_mapping_expected_value,
            mapping_optional={},
            data_frame=df_expected_value,
            scope="default",
            file_type="instruments",
            identifier_mapping=identifier_mapping_expected_value,
        )

        result = cocoon.instruments.resolve_instruments(
            self.api_factory, data_frame=df, identifier_mapping=identifier_mapping
        )
        result.set_index("ids", inplace=True)

        def validate_luid(luid):
            self.assertTrue(luid.startswith("LUID_"), f"unexpected LUID value: {luid}")

        for i in source_ids:
            validate_luid(result.loc[i]["LusidInstrumentId"])


    def test_resolve_invalid_instrument(self):
        data = {"ids": ["blah"], "is_cash": [False]}

        identifier_mapping = {"Figi": "ids", "Currency": "is_cash"}

        df = pd.DataFrame(data)

        result = cocoon.instruments.resolve_instruments(
            self.api_factory, data_frame=df, identifier_mapping=identifier_mapping
        )

        self.assertFalse(result.iloc[0]["LusidInstrumentId"])

    @parameterized.expand(
        [
            [
                "Standard Out of Box Identifiers",
                [
                    "Figi",
                    "ClientInternal",
                    "QuotePermId",
                    "LusidInstrumentId",
                    "Currency",
                ],
            ]
        ]
    )
    @unittest.skip("Currency holdings return a message instead of LUID-XXXXXX")
    def test_get_unique_identifiers(self, _, expected_outcome) -> None:
        """
        Tests that the unique identifers are correctly retrieved

        :param expected_outcome: The expected outcome

        :return: None
        """
        unique_identifiers = cocoon.instruments.get_unique_identifiers(self.api_factory)
        unique_identifiers.sort()
        expected_outcome.sort()

        self.assertListEqual(unique_identifiers, expected_outcome)

    @parameterized.expand(
        [
            [
                "A standard successful load of instruments",
                "data/global-fund-combined-instrument-master-missing-name.csv",
                {"name": "instrument_name"},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
            ],
            [
                "Constant prefix",
                "data/global-fund-combined-instrument-master-missing-name.csv",
                {"name": "$Unknown"},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
            ],
            [
                "Default specified",
                "data/global-fund-combined-instrument-master-missing-name.csv",
                {"name": {"default": "$Unknown"}},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
            ],
            [
                "Default specified with column name",
                "data/global-fund-combined-instrument-master-missing-name.csv",
                {"name": {"default": "$Unknown", "column": "instrument_name"}},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
            ],
        ]
    )
    @unittest.skip
    def test_enrich_instruments(
        self, _, file_name, mapping_required, instrument_identifier_mapping
    ):

        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        data_frame_updated, mapping_required = asyncio.run_coroutine_threadsafe(
            cocoon.instruments.enrich_instruments(
                api_factory=self.api_factory,
                data_frame=data_frame,
                instrument_identifier_mapping=instrument_identifier_mapping,
                mapping_required=mapping_required,
                constant_prefix="$",
            ),
            self.loop,
        ).result()



import io
import unittest
from unittest.mock import patch

import pandas

import finbourne_sdk_utils.lpt.dfq as dfq
from pandas.testing import assert_frame_equal


class DfqTests(unittest.TestCase):
    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_not_df(self, mock_stdout):
        with self.assertRaises(SystemExit):
            args = dfq.parse(False, ["-g", "Name"])
            dfq.dfq(args, "No reconciliation breaks")
            self.assertEqual("No reconciliation breaks", mock_stdout.getvalue().strip())

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_first(self, mock_stdout):
        args = dfq.parse(False, ["-f", "2"])
        dfq.dfq(
            args,
            pandas.DataFrame(
                [
                    {"Name": "Foo", "Value": 2},
                    {"Name": "Foo", "Value": 5},
                    {"Name": "Bar", "Value": 7},
                ]
            ),
        )

        expected_df = pandas.DataFrame(
            [{"Name": "Foo", "Value": 2}, {"Name": "Foo", "Value": 5},]
        ).to_string(index=False)
        expected = f"First 2\n{expected_df}"

        self.assertEqual(expected, mock_stdout.getvalue().strip())

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_last(self, mock_stdout):
        args = dfq.parse(False, ["-l", "1"])
        dfq.dfq(
            args,
            pandas.DataFrame(
                [
                    {"Name": "Foo", "Value": 2},
                    {"Name": "Foo", "Value": 5},
                    {"Name": "Bar", "Value": 7},
                ]
            ),
        )

        expected_df = pandas.DataFrame([{"Name": "Bar", "Value": 7},]).to_string(
            index=False
        )
        expected = f"Last 1\n{expected_df}"

        self.assertEqual(expected, mock_stdout.getvalue().strip())

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_first_last_both_0(self, mock_stdout):
        args = dfq.parse(False, ["-l", "0", "-f", "0"])
        dfq.dfq(
            args,
            pandas.DataFrame(
                [
                    {"Name": "Foo", "Value": 2},
                    {"Name": "Foo", "Value": 5},
                    {"Name": "Bar", "Value": 7},
                ]
            ),
        )

        expected_df = pandas.DataFrame(
            [
                {"Name": "Foo", "Value": 2},
                {"Name": "Foo", "Value": 5},
                {"Name": "Bar", "Value": 7},
            ]
        ).to_string(index=False)
        expected = f"{expected_df}"

        self.assertEqual(expected, mock_stdout.getvalue().strip())

    def test_group_by(self):
        args = dfq.parse(False, ["-g", "Name"])
        result = dfq.apply_args(
            args,
            pandas.DataFrame(
                [
                    {"Name": "Foo", "Value": 2},
                    {"Name": "Foo", "Value": 5},
                    {"Name": "Bar", "Value": 7},
                ]
            ),
        )

        assert_frame_equal(
            pandas.DataFrame(
                [{"Name": "Bar", "Value": 7}, {"Name": "Foo", "Value": 7},]
            ),
            result,
        )

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_columns(self, mock_stdout):
        args = dfq.parse(False, ["--columns"])

        with self.assertRaises(SystemExit):
            dfq.dfq(args, pandas.DataFrame([{"Name": "Foo", "Value": 2},]))

            expected = "Name\nValue"

            self.assertEqual(expected, mock_stdout.getvalue().strip())

    def test_unique(self):
        args = dfq.parse(False, ["-u"])
        result = dfq.apply_args(
            args,
            pandas.DataFrame(
                [{"Name": "Foo", "Value": 2}, {"Name": "Foo", "Value": 2}]
            ),
        )

        assert_frame_equal(pandas.DataFrame([{"Name": "Foo", "Value": 2}]), result)

    def test_where(self):
        args = dfq.parse(False, ["--where", "Value>3"])

        result = dfq.apply_args(
            args,
            pandas.DataFrame(
                [
                    {"Name": "Foo", "Value": 2},
                    {"Name": "Bar", "Value": 4},
                    {"Name": "Baz", "Value": 10},
                ]
            ),
        )

        expected_df = pandas.DataFrame(
            [{"Name": "Bar", "Value": 4}, {"Name": "Baz", "Value": 10},]
        )

        assert_frame_equal(
            expected_df.reset_index(drop=True), result.reset_index(drop=True)
        )

    def test_where_multiple_conditions(self):
        args = dfq.parse(False, ["--where", "Value>3", "Name=Baz"])

        result = dfq.apply_args(
            args,
            pandas.DataFrame(
                [
                    {"Name": "Foo", "Value": 2},
                    {"Name": "Bar", "Value": 4},
                    {"Name": "Baz", "Value": 10},
                ]
            ),
        )

        assert_frame_equal(
            pandas.DataFrame([{"Name": "Baz", "Value": 10},]).reset_index(drop=True),
            result.reset_index(drop=True),
        )

    def test_order(self):
        args = dfq.parse(False, ["--order", "Name"])

        result = dfq.apply_args(
            args,
            pandas.DataFrame(
                [
                    {"Name": "Bar", "Value": 2},
                    {"Name": "Foo", "Value": 4},
                    {"Name": "Baz", "Value": 10},
                ]
            ),
        )

        expected_df = pandas.DataFrame(
            [
                {"Name": "Bar", "Value": 2},
                {"Name": "Baz", "Value": 10},
                {"Name": "Foo", "Value": 4},
            ]
        )

        assert_frame_equal(
            expected_df.reset_index(drop=True), result.reset_index(drop=True)
        )

    def test_select(self):
        args = dfq.parse(False, ["--select", "Name", "Value1"])

        result = dfq.apply_args(
            args,
            pandas.DataFrame(
                [
                    {"Name": "Bar", "Value1": 2, "Value2": 4},
                    {"Name": "Foo", "Value1": 4, "Value2": 6},
                    {"Name": "Baz", "Value1": 10, "Value2": 8},
                ]
            ),
        )

        expected_df = pandas.DataFrame(
            [
                {"Name": "Bar", "Value1": 2},
                {"Name": "Foo", "Value1": 4},
                {"Name": "Baz", "Value1": 10},
            ]
        )

        assert_frame_equal(
            expected_df.reset_index(drop=True), result.reset_index(drop=True)
        )
